import json
from unittest import IsolatedAsyncioTestCase

import pytest

from .....admin.request_context import AdminRequestContext
from .....connections.models.conn_record import ConnRecord
from .....ledger.base import BaseLedger
from .....tests import mock
from .....utils.testing import create_test_profile
from .....wallet.base import BaseWallet
from .....wallet.did_info import DIDInfo
from .....wallet.did_method import SOV
from .....wallet.key_type import ED25519
from .. import routes as test_module
from ..models.transaction_record import TransactionRecord

TEST_DID = "LjgpST2rjsoxYegQDRm7EL"
SCHEMA_NAME = "bc-reg"
SCHEMA_TXN = 12
SCHEMA_ID = f"{TEST_DID}:2:{SCHEMA_NAME}:1.0"
CRED_DEF_ID = f"{TEST_DID}:3:CL:12:tag1"


class TestEndorseTransactionRoutes(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "secret-key",
            }
        )
        self.context = AdminRequestContext.test_context({}, self.profile)

        self.ledger = mock.MagicMock(BaseLedger, autospec=True)
        self.ledger.txn_submit = mock.CoroutineMock(
            return_value=json.dumps(
                {
                    "result": {
                        "txn": {"type": "101", "metadata": {"from": TEST_DID}},
                        "txnMetadata": {"txnId": SCHEMA_ID},
                    }
                }
            )
        )
        self.ledger.get_schema = mock.CoroutineMock(
            return_value={"id": SCHEMA_ID, "...": "..."}
        )
        self.profile.context.injector.bind_instance(BaseLedger, self.ledger)

        self.request_dict = {
            "context": self.context,
            "outbound_message_router": mock.CoroutineMock(),
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "secret-key"},
        )

        self.test_did = "sample-did"

    async def test_transactions_list(self):
        with (
            mock.patch.object(
                TransactionRecord, "query", mock.CoroutineMock()
            ) as mock_query,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_query.return_value = [
                mock.MagicMock(serialize=mock.MagicMock(return_value={"...": "..."}))
            ]
            await test_module.transactions_list(self.request)

            mock_response.assert_called_once_with({"results": [{"...": "..."}]})

    async def test_transactions_list_x(self):
        with (
            mock.patch.object(
                TransactionRecord, "query", mock.CoroutineMock()
            ) as mock_query,
            mock.patch.object(test_module.web, "json_response"),
        ):
            mock_query.side_effect = test_module.StorageError()

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.transactions_list(self.request)

    async def test_transactions_retrieve(self):
        self.request.match_info = {"tran_id": "dummy"}

        with (
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_retrieve,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )
            await test_module.transactions_retrieve(self.request)

            mock_response.assert_called_once_with({"...": "..."})

    async def test_transactions_retrieve_not_found_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        with mock.patch.object(
            TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_retrieve:
            mock_retrieve.side_effect = test_module.StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.transactions_retrieve(self.request)

    async def test_transactions_retrieve_base_model_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        with mock.patch.object(
            TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_retrieve:
            mock_retrieve.side_effect = test_module.BaseModelError()

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.transactions_retrieve(self.request)

    async def test_transaction_create_request(self):
        self.request.query = {
            "tran_id": "dummy",
        }
        self.request.json = mock.CoroutineMock(
            return_value={
                "expires_time": "2021-03-29T05:22:19Z",
            }
        )
        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                create_request=mock.CoroutineMock(
                    return_value=(
                        mock.MagicMock(
                            serialize=mock.MagicMock(return_value={"...": "..."})
                        ),
                        mock.MagicMock(),
                    )
                )
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_my_job": (
                            test_module.TransactionJob.TRANSACTION_AUTHOR.name
                        ),
                        "transaction_their_job": (
                            test_module.TransactionJob.TRANSACTION_ENDORSER.name
                        ),
                    }
                )
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )
            await test_module.transaction_create_request(self.request)

            mock_response.assert_called_once_with({"...": "..."})

    async def test_transaction_create_request_not_found_x(self):
        self.request.query = {
            "tran_id": "dummy",
        }
        self.request.json = mock.CoroutineMock(
            return_value={
                "expires_time": "2021-03-29T05:22:19Z",
            }
        )
        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve:
            mock_conn_rec_retrieve.side_effect = test_module.StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.transaction_create_request(self.request)

    async def test_transaction_create_request_base_model_x(self):
        self.request.query = {
            "tran_id": "dummy",
        }
        self.request.json = mock.CoroutineMock(
            return_value={
                "expires_time": "2021-03-29T05:22:19Z",
            }
        )
        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
        ):
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_my_job": (
                            test_module.TransactionJob.TRANSACTION_AUTHOR.name
                        ),
                        "transaction_their_job": (
                            test_module.TransactionJob.TRANSACTION_ENDORSER.name
                        ),
                    }
                )
            )
            mock_txn_rec_retrieve.side_effect = test_module.BaseModelError()

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.transaction_create_request(self.request)

    async def test_transaction_create_request_no_jobs_x(self):
        self.request.query = {
            "tran_id": "dummy",
        }
        self.request.json = mock.CoroutineMock(
            return_value={
                "expires_time": "2021-03-29T05:22:19Z",
            }
        )
        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                create_request=mock.CoroutineMock(
                    return_value=(
                        mock.MagicMock(
                            serialize=mock.MagicMock(return_value={"...": "..."})
                        ),
                        mock.MagicMock(),
                    )
                )
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(return_value=None)
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.transaction_create_request(self.request)

    async def test_transaction_create_request_no_my_job_x(self):
        self.request.query = {
            "tran_id": "dummy",
        }
        self.request.json = mock.CoroutineMock(
            return_value={
                "expires_time": "2021-03-29T05:22:19Z",
            }
        )
        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                create_request=mock.CoroutineMock(
                    return_value=(
                        mock.MagicMock(
                            serialize=mock.MagicMock(return_value={"...": "..."})
                        ),
                        mock.MagicMock(),
                    )
                )
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_their_job": (
                            test_module.TransactionJob.TRANSACTION_ENDORSER.name
                        ),
                    }
                )
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.transaction_create_request(self.request)

    async def test_transaction_create_request_no_their_job_x(self):
        self.request.query = {
            "tran_id": "dummy",
        }
        self.request.json = mock.CoroutineMock(
            return_value={
                "expires_time": "2021-03-29T05:22:19Z",
            }
        )
        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                create_request=mock.CoroutineMock(
                    return_value=(
                        mock.MagicMock(
                            serialize=mock.MagicMock(return_value={"...": "..."})
                        ),
                        mock.MagicMock(),
                    )
                )
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_my_job": (
                            test_module.TransactionJob.TRANSACTION_AUTHOR.name
                        ),
                    }
                )
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.transaction_create_request(self.request)

    async def test_transaction_create_request_my_wrong_job_x(self):
        self.request.query = {
            "tran_id": "dummy",
        }
        self.request.json = mock.CoroutineMock(
            return_value={
                "expires_time": "2021-03-29T05:22:19Z",
            }
        )
        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
        ):
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_their_job": (
                            test_module.TransactionJob.TRANSACTION_ENDORSER.name
                        ),
                        "transaction_my_job": "a suffusion of yellow",
                    }
                )
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.transaction_create_request(self.request)

    async def test_transaction_create_request_mgr_create_request_x(self):
        self.request.query = {
            "tran_id": "dummy",
        }
        self.request.json = mock.CoroutineMock(
            return_value={
                "expires_time": "2021-03-29T05:22:19Z",
            }
        )
        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                create_request=mock.CoroutineMock(
                    side_effect=test_module.TransactionManagerError()
                )
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_my_job": (
                            test_module.TransactionJob.TRANSACTION_AUTHOR.name
                        ),
                        "transaction_their_job": (
                            test_module.TransactionJob.TRANSACTION_ENDORSER.name
                        ),
                    }
                )
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.transaction_create_request(self.request)

    async def test_endorse_transaction_response(self):
        self.request.match_info = {"tran_id": "dummy"}
        self.profile.context.injector.bind_instance(
            BaseWallet,
            mock.MagicMock(
                get_public_did=mock.CoroutineMock(
                    return_value=DIDInfo(
                        "did",
                        "verkey",
                        {"meta": "data"},
                        method=SOV,
                        key_type=ED25519,
                    )
                )
            ),
        )

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                create_endorse_response=mock.CoroutineMock(
                    return_value=(
                        mock.MagicMock(
                            serialize=mock.MagicMock(return_value={"...": "..."})
                        ),
                        mock.MagicMock(),
                    )
                )
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_my_job": (
                            test_module.TransactionJob.TRANSACTION_ENDORSER.name
                        )
                    }
                )
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )
            await test_module.endorse_transaction_response(self.request)

            mock_response.assert_called_once_with({"...": "..."})

    # TODO code re-factored from routes.py to manager.py so tests must be moved
    @pytest.mark.skip("Need to fix")
    async def test_endorse_transaction_response_no_wallet_x(self):
        self.session.context.injector.clear_binding(BaseWallet)
        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.endorse_transaction_response(self.request)

    @pytest.mark.skip("Need to fix")
    async def test_endorse_transaction_response_no_endorser_did_info_x(self):
        self.request.match_info = {"tran_id": "dummy"}
        self.session.context.injector.bind_instance(
            BaseWallet,
            mock.MagicMock(get_public_did=mock.CoroutineMock(return_value=None)),
        )
        with mock.patch.object(
            self.context.profile,
            "session",
            mock.MagicMock(return_value=self.session),
        ):
            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.endorse_transaction_response(self.request)

    async def test_endorse_transaction_response_not_found_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        self.profile.context.injector.bind_instance(
            BaseWallet,
            mock.MagicMock(
                get_public_did=mock.CoroutineMock(
                    return_value=DIDInfo(
                        "did",
                        "verkey",
                        {"meta": "data"},
                        method=SOV,
                        key_type=ED25519,
                    )
                )
            ),
        )

        with mock.patch.object(
            TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_txn_rec_retrieve:
            mock_txn_rec_retrieve.side_effect = test_module.StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.endorse_transaction_response(self.request)

    async def test_endorse_transaction_response_base_model_x(self):
        self.request.match_info = {"tran_id": "dummy"}
        self.profile.context.injector.bind_instance(
            BaseWallet,
            mock.MagicMock(
                get_public_did=mock.CoroutineMock(
                    return_value=DIDInfo(
                        "did",
                        "verkey",
                        {"meta": "data"},
                        method=SOV,
                        key_type=ED25519,
                    )
                )
            ),
        )

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
        ):
            mock_conn_rec_retrieve.side_effect = test_module.BaseModelError()
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.endorse_transaction_response(self.request)

    async def test_endorse_transaction_response_no_jobs_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        self.profile.context.injector.bind_instance(
            BaseWallet,
            mock.MagicMock(
                get_public_did=mock.CoroutineMock(
                    return_value=DIDInfo(
                        "did",
                        "verkey",
                        {"meta": "data"},
                        method=SOV,
                        key_type=ED25519,
                    )
                )
            ),
        )

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
        ):
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(return_value=None)
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.endorse_transaction_response(self.request)

    @pytest.mark.skip("Need to fix")
    async def test_endorse_transaction_response_no_ledger_x(self):
        self.request.match_info = {"tran_id": "dummy"}
        self.context.injector.clear_binding(BaseLedger)
        self.profile.context.injector.bind_instance(
            BaseWallet,
            mock.MagicMock(
                get_public_did=mock.CoroutineMock(
                    return_value=DIDInfo(
                        "did",
                        "verkey",
                        {"meta": "data"},
                        method=SOV,
                        key_type=ED25519,
                    )
                )
            ),
        )

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                create_endorse_response=mock.CoroutineMock(
                    return_value=(
                        mock.MagicMock(
                            serialize=mock.MagicMock(return_value={"...": "..."})
                        ),
                        mock.MagicMock(),
                    )
                )
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_my_job": (
                            test_module.TransactionJob.TRANSACTION_ENDORSER.name
                        )
                    }
                )
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.endorse_transaction_response(self.request)

    async def test_endorse_transaction_response_wrong_my_job_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        self.profile.context.injector.bind_instance(
            BaseWallet,
            mock.MagicMock(
                get_public_did=mock.CoroutineMock(
                    return_value=DIDInfo(
                        "did",
                        "verkey",
                        {"meta": "data"},
                        method=SOV,
                        key_type=ED25519,
                    )
                )
            ),
        )

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
        ):
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_my_job": (
                            test_module.TransactionJob.TRANSACTION_AUTHOR.name
                        )
                    }
                )
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.endorse_transaction_response(self.request)

    @pytest.mark.skip("Need to fix")
    async def test_endorse_transaction_response_ledger_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        self.profile.context.injector.bind_instance(
            BaseWallet,
            mock.MagicMock(
                get_public_did=mock.CoroutineMock(
                    return_value=DIDInfo(
                        "did",
                        "verkey",
                        {"meta": "data"},
                        method=SOV,
                        key_type=ED25519,
                    )
                )
            ),
        )
        self.ledger.txn_endorse = mock.CoroutineMock(
            side_effect=test_module.LedgerError()
        )

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                create_endorse_response=mock.CoroutineMock(
                    return_value=(
                        mock.MagicMock(
                            serialize=mock.MagicMock(return_value={"...": "..."})
                        ),
                        mock.MagicMock(),
                    )
                )
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_my_job": (
                            test_module.TransactionJob.TRANSACTION_ENDORSER.name
                        )
                    }
                )
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.endorse_transaction_response(self.request)

    async def test_endorse_transaction_response_txn_mgr_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        self.profile.context.injector.bind_instance(
            BaseWallet,
            mock.MagicMock(
                get_public_did=mock.CoroutineMock(
                    return_value=DIDInfo(
                        "did",
                        "verkey",
                        {"meta": "data"},
                        method=SOV,
                        key_type=ED25519,
                    )
                )
            ),
        )

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
            mock.patch.object(test_module.web, "json_response"),
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                create_endorse_response=mock.CoroutineMock(
                    side_effect=test_module.TransactionManagerError()
                )
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_my_job": (
                            test_module.TransactionJob.TRANSACTION_ENDORSER.name
                        )
                    }
                )
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.endorse_transaction_response(self.request)

    async def test_refuse_transaction_response(self):
        self.request.match_info = {"tran_id": "dummy"}

        self.profile.context.injector.bind_instance(
            BaseWallet,
            mock.MagicMock(
                get_public_did=mock.CoroutineMock(
                    return_value=DIDInfo(
                        "did",
                        "verkey",
                        {"meta": "data"},
                        method=SOV,
                        key_type=ED25519,
                    )
                )
            ),
        )

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                create_refuse_response=mock.CoroutineMock(
                    return_value=(
                        mock.MagicMock(  # transaction
                            connection_id="dummy",
                            serialize=mock.MagicMock(return_value={"...": "..."}),
                        ),
                        mock.MagicMock(),  # refused_transaction_response
                    )
                )
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_my_job": (
                            test_module.TransactionJob.TRANSACTION_ENDORSER.name
                        )
                    }
                )
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )
            await test_module.refuse_transaction_response(self.request)

            mock_response.assert_called_once_with({"...": "..."})

    async def test_refuse_transaction_response_not_found_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        self.profile.context.injector.bind_instance(
            BaseWallet,
            mock.MagicMock(
                get_public_did=mock.CoroutineMock(
                    return_value=DIDInfo(
                        "did",
                        "verkey",
                        {"meta": "data"},
                        method=SOV,
                        key_type=ED25519,
                    )
                )
            ),
        )

        with mock.patch.object(
            TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_txn_rec_retrieve:
            mock_txn_rec_retrieve.side_effect = test_module.StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.refuse_transaction_response(self.request)

    async def test_refuse_transaction_response_conn_base_model_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        self.profile.context.injector.bind_instance(
            BaseWallet,
            mock.MagicMock(
                get_public_did=mock.CoroutineMock(
                    return_value=DIDInfo(
                        "did",
                        "verkey",
                        {"meta": "data"},
                        method=SOV,
                        key_type=ED25519,
                    )
                )
            ),
        )

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
        ):
            mock_conn_rec_retrieve.side_effect = test_module.BaseModelError()
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.refuse_transaction_response(self.request)

    async def test_refuse_transaction_response_no_jobs_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        self.profile.context.injector.bind_instance(
            BaseWallet,
            mock.MagicMock(
                get_public_did=mock.CoroutineMock(
                    return_value=DIDInfo(
                        "did",
                        "verkey",
                        {"meta": "data"},
                        method=SOV,
                        key_type=ED25519,
                    )
                )
            ),
        )

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
        ):
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(return_value=None)
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.refuse_transaction_response(self.request)

    async def test_refuse_transaction_response_wrong_my_job_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        self.profile.context.injector.bind_instance(
            BaseWallet,
            mock.MagicMock(
                get_public_did=mock.CoroutineMock(
                    return_value=DIDInfo(
                        "did",
                        "verkey",
                        {"meta": "data"},
                        method=SOV,
                        key_type=ED25519,
                    )
                )
            ),
        )

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
        ):
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_my_job": (
                            test_module.TransactionJob.TRANSACTION_AUTHOR.name
                        )
                    }
                )
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.refuse_transaction_response(self.request)

    async def test_refuse_transaction_response_txn_mgr_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        self.profile.context.injector.bind_instance(
            BaseWallet,
            mock.MagicMock(
                get_public_did=mock.CoroutineMock(
                    return_value=DIDInfo(
                        "did",
                        "verkey",
                        {"meta": "data"},
                        method=SOV,
                        key_type=ED25519,
                    )
                )
            ),
        )

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
            mock.patch.object(test_module.web, "json_response"),
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                create_refuse_response=mock.CoroutineMock(
                    side_effect=test_module.TransactionManagerError()
                )
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_my_job": (
                            test_module.TransactionJob.TRANSACTION_ENDORSER.name
                        )
                    }
                )
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.refuse_transaction_response(self.request)

    async def test_cancel_transaction(self):
        self.request.match_info = {"tran_id": "dummy"}

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                cancel_transaction=mock.CoroutineMock(
                    return_value=(
                        mock.MagicMock(  # transaction
                            connection_id="dummy",
                            serialize=mock.MagicMock(return_value={"...": "..."}),
                        ),
                        mock.MagicMock(),  # refused_transaction_response
                    )
                )
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_my_job": (
                            test_module.TransactionJob.TRANSACTION_AUTHOR.name
                        )
                    }
                )
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )
            await test_module.cancel_transaction(self.request)

            mock_response.assert_called_once_with({"...": "..."})

    async def test_cancel_transaction_not_found_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        with mock.patch.object(
            TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_txn_rec_retrieve:
            mock_txn_rec_retrieve.side_effect = test_module.StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.cancel_transaction(self.request)

    async def test_cancel_transaction_conn_rec_base_model_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
        ):
            mock_conn_rec_retrieve.side_effect = test_module.BaseModelError()
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.cancel_transaction(self.request)

    async def test_cancel_transaction_no_jobs_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
        ):
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(return_value=None)
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.cancel_transaction(self.request)

    async def test_cancel_transaction_wrong_my_job_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
        ):
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_my_job": (
                            test_module.TransactionJob.TRANSACTION_ENDORSER.name
                        )
                    }
                )
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.cancel_transaction(self.request)

    async def test_cancel_transaction_txn_mgr_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
            mock.patch.object(test_module.web, "json_response"),
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                cancel_transaction=mock.CoroutineMock(
                    side_effect=test_module.TransactionManagerError()
                )
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_my_job": (
                            test_module.TransactionJob.TRANSACTION_AUTHOR.name
                        )
                    }
                )
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.cancel_transaction(self.request)

    async def test_transaction_resend(self):
        self.request.match_info = {"tran_id": "dummy"}

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                transaction_resend=mock.CoroutineMock(
                    return_value=(
                        mock.MagicMock(  # transaction
                            connection_id="dummy",
                            serialize=mock.MagicMock(return_value={"...": "..."}),
                        ),
                        mock.MagicMock(),  # refused_transaction_response
                    )
                )
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_my_job": (
                            test_module.TransactionJob.TRANSACTION_AUTHOR.name
                        )
                    }
                )
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )
            await test_module.transaction_resend(self.request)

        mock_response.assert_called_once_with({"...": "..."})

    async def test_transaction_resend_not_found_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        with mock.patch.object(
            TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_txn_rec_retrieve:
            mock_txn_rec_retrieve.side_effect = test_module.StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.transaction_resend(self.request)

    async def test_transaction_resend_conn_rec_base_model_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
        ):
            mock_conn_rec_retrieve.side_effect = test_module.BaseModelError()
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.transaction_resend(self.request)

    async def test_transaction_resend_no_jobs_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
        ):
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(return_value=None)
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.transaction_resend(self.request)

    async def test_transaction_resend_my_wrong_job_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
        ):
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_their_job": (
                            test_module.TransactionJob.TRANSACTION_ENDORSER.name
                        ),
                        "transaction_my_job": "a suffusion of yellow",
                    }
                )
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.transaction_resend(self.request)

    async def test_transaction_resend_txn_mgr_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
            mock.patch.object(test_module.web, "json_response"),
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                transaction_resend=mock.CoroutineMock(
                    side_effect=test_module.TransactionManagerError()
                )
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_my_job": (
                            test_module.TransactionJob.TRANSACTION_AUTHOR.name
                        )
                    }
                )
            )
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."})
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.transaction_resend(self.request)

    async def test_set_endorser_role(self):
        self.request.match_info = {"conn_id": "dummy"}

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                set_transaction_my_job=mock.CoroutineMock()
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_my_job": (
                            test_module.TransactionJob.TRANSACTION_AUTHOR.name
                        )
                    }
                )
            )
            await test_module.set_endorser_role(self.request)

        mock_response.assert_called_once_with(
            {"transaction_my_job": test_module.TransactionJob.TRANSACTION_AUTHOR.name}
        )

    async def test_set_endorser_role_not_found_x(self):
        self.request.match_info = {"conn_id": "dummy"}

        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve:
            mock_conn_rec_retrieve.side_effect = test_module.StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.set_endorser_role(self.request)

    async def test_set_endorser_role_base_model_x(self):
        self.request.match_info = {"conn_id": "dummy"}

        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve:
            mock_conn_rec_retrieve.side_effect = test_module.BaseModelError()

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.set_endorser_role(self.request)

    async def test_set_endorser_info(self):
        self.request.match_info = {"conn_id": "dummy"}
        self.request.query = {"endorser_did": "did", "endorser_name": "name"}
        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_my_job": (
                            test_module.TransactionJob.TRANSACTION_AUTHOR.name
                        ),
                        "transaction_their_job": (
                            test_module.TransactionJob.TRANSACTION_ENDORSER.name
                        ),
                    }
                ),
                metadata_set=mock.CoroutineMock(),
            )
            await test_module.set_endorser_info(self.request)

            mock_response.assert_called_once_with(
                {
                    "transaction_my_job": "TRANSACTION_AUTHOR",
                    "transaction_their_job": "TRANSACTION_ENDORSER",
                    "endorser_did": "did",
                    "endorser_name": "name",
                }
            )

    async def test_set_endorser_info_no_prior_value(self):
        self.request.match_info = {"conn_id": "dummy"}
        self.request.query = {"endorser_did": "did", "endorser_name": "name"}
        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    side_effect=[
                        {
                            "transaction_my_job": (
                                test_module.TransactionJob.TRANSACTION_AUTHOR.name
                            ),
                            "transaction_their_job": (
                                test_module.TransactionJob.TRANSACTION_ENDORSER.name
                            ),
                        },
                        None,
                        {
                            "endorser_did": "did",
                            "endorser_name": "name",
                        },
                    ]
                ),
                metadata_set=mock.CoroutineMock(),
            )
            await test_module.set_endorser_info(self.request)

            mock_response.assert_called_once_with(
                {
                    "endorser_did": "did",
                    "endorser_name": "name",
                }
            )

    async def test_set_endorser_info_not_found_x(self):
        self.request.match_info = {"conn_id": "dummy"}
        self.request.query = {"endorser_did": "did", "endorser_name": "name"}

        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve:
            mock_conn_rec_retrieve.side_effect = test_module.StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.set_endorser_info(self.request)

    async def test_set_endorser_info_base_model_x(self):
        self.request.match_info = {"conn_id": "dummy"}
        self.request.query = {"endorser_did": "did", "endorser_name": "name"}

        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve:
            mock_conn_rec_retrieve.side_effect = test_module.BaseModelError()

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.set_endorser_info(self.request)

    async def test_set_endorser_info_no_transaction_jobs_x(self):
        self.request.match_info = {"conn_id": "dummy"}
        self.request.query = {"endorser_did": "did", "endorser_name": "name"}

        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve:
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(return_value=None)
            )
            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.set_endorser_info(self.request)

    async def test_set_endorser_info_no_transaction_my_job_x(self):
        self.request.match_info = {"conn_id": "dummy"}
        self.request.query = {"endorser_did": "did", "endorser_name": "name"}

        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve:
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_their_job": (
                            test_module.TransactionJob.TRANSACTION_ENDORSER.name
                        ),
                    }
                )
            )
            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.set_endorser_info(self.request)

    async def test_set_endorser_info_my_wrong_job_x(self):
        self.request.match_info = {"conn_id": "dummy"}
        self.request.query = {"endorser_did": "did", "endorser_name": "name"}
        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve:
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "transaction_their_job": (
                            test_module.TransactionJob.TRANSACTION_ENDORSER.name
                        ),
                        "transaction_my_job": "a suffusion of yellow",
                    }
                )
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.set_endorser_info(self.request)

    async def test_transaction_write_schema_txn(self):
        self.request.match_info = {"tran_id": "dummy"}
        with (
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_txn_mgr.return_value.complete_transaction = mock.CoroutineMock()

            mock_txn_mgr.return_value.complete_transaction.return_value = (
                mock.CoroutineMock(serialize=mock.MagicMock(return_value={"...": "..."})),
                mock.CoroutineMock(),
            )

            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(),
                state=TransactionRecord.STATE_TRANSACTION_ENDORSED,
                messages_attach=[{"data": {"json": json.dumps({"message": "attached"})}}],
            )
            await test_module.transaction_write(self.request)
            mock_response.assert_called_once_with({"...": "..."})

    async def test_transaction_write_not_found_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        with mock.patch.object(
            TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_txn_rec_retrieve:
            mock_txn_rec_retrieve.side_effect = test_module.StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.transaction_write(self.request)

    async def test_transaction_write_base_model_x(self):
        self.request.match_info = {"tran_id": "dummy"}

        with mock.patch.object(
            TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_txn_rec_retrieve:
            mock_txn_rec_retrieve.side_effect = test_module.BaseModelError()

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.transaction_write(self.request)

    async def test_transaction_write_wrong_state_x(self):
        self.request.match_info = {"tran_id": "dummy"}
        with mock.patch.object(
            TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_txn_rec_retrieve:
            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."}),
                state=TransactionRecord.STATE_TRANSACTION_CREATED,
                messages_attach=[{"data": {"json": json.dumps({"message": "attached"})}}],
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.transaction_write(self.request)

    async def test_transaction_write_schema_txn_complete_x(self):
        self.request.match_info = {"tran_id": "dummy"}
        with (
            mock.patch.object(
                TransactionRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_txn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                complete_transaction=mock.CoroutineMock(
                    side_effect=test_module.StorageError()
                )
            )

            mock_txn_rec_retrieve.return_value = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"...": "..."}),
                state=TransactionRecord.STATE_TRANSACTION_ENDORSED,
                messages_attach=[{"data": {"json": json.dumps({"message": "attached"})}}],
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.transaction_write(self.request)

    async def test_register(self):
        mock_app = mock.MagicMock()
        mock_app.add_routes = mock.MagicMock()

        await test_module.register(mock_app)
        mock_app.add_routes.assert_called_once()

    async def test_post_process_routes(self):
        mock_app = mock.MagicMock(_state={"swagger_dict": {"paths": {}}})
        test_module.post_process_routes(mock_app)

        assert "tags" in mock_app._state["swagger_dict"]



"""Test connections base manager."""

import secrets
from unittest import IsolatedAsyncioTestCase
from unittest.mock import call

import base58
import pytest
from pydid import DID, DIDDocument, DIDDocumentBuilder
from pydid.doc.builder import ServiceBuilder
from pydid.verification_method import (
    EcdsaSecp256k1VerificationKey2019,
    Ed25519VerificationKey2018,
    Ed25519VerificationKey2020,
    JsonWebKey2020,
)

from ...cache.base import BaseCache
from ...cache.in_memory import InMemoryCache
from ...config.base import InjectionError
from ...connections.base_manager import BaseConnectionManagerError
from ...connections.models.conn_record import ConnRecord
from ...connections.models.connection_target import ConnectionTarget
from ...connections.models.diddoc import DIDDoc, PublicKey, PublicKeyType, Service
from ...core.oob_processor import OobMessageProcessor
from ...did.did_key import DIDKey
from ...messaging.responder import BaseResponder, MockResponder
from ...multitenant.base import BaseMultitenantManager
from ...multitenant.manager import MultitenantManager
from ...protocols.coordinate_mediation.v1_0.route_manager import (
    CoordinateMediationV1RouteManager,
    RouteManager,
)
from ...protocols.discovery.v2_0.manager import V20DiscoveryMgr
from ...protocols.out_of_band.v1_0.messages.invitation import InvitationMessage
from ...protocols.out_of_band.v1_0.messages.service import Service as OOBService
from ...resolver.default.key import KeyDIDResolver
from ...resolver.default.legacy_peer import LegacyPeerDIDResolver
from ...resolver.default.peer4 import PeerDID4Resolver
from ...resolver.did_resolver import DIDResolver
from ...storage.error import StorageNotFoundError
from ...tests import mock
from ...transport.inbound.receipt import MessageReceipt
from ...utils.multiformats import multibase, multicodec
from ...utils.testing import create_test_profile
from ...wallet.askar import AskarWallet
from ...wallet.base import BaseWallet, DIDInfo
from ...wallet.did_method import SOV, DIDMethods
from ...wallet.error import WalletNotFoundError
from ...wallet.key_type import ED25519, KeyTypes
from ...wallet.util import b58_to_bytes, bytes_to_b64
from ..base_manager import BaseConnectionManager


class TestBaseConnectionManager(IsolatedAsyncioTestCase):
    def make_did_doc(self, did, verkey):
        doc = DIDDoc(did=did)
        controller = did
        ident = "1"
        pk_value = verkey
        pk = PublicKey(
            did, ident, pk_value, PublicKeyType.ED25519_SIG_2018, controller, False
        )
        doc.set(pk)
        recip_keys = [pk]
        router_keys = []
        service = Service(
            did, "indy", "IndyAgent", recip_keys, router_keys, self.test_endpoint
        )
        doc.set(service)
        return doc

    async def asyncSetUp(self):
        self.test_seed = "testseed000000000000000000000001"
        self.test_did = "55GkHamhTU1ZbTbV2ab9DE"
        self.test_verkey = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
        self.test_endpoint = "http://localhost"

        self.test_target_did = "GbuDUYXaUZRfHD2jeDuQuP"
        self.test_target_verkey = "9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"

        self.test_pthid = "test-pthid"

        self.responder = MockResponder()

        self.oob_mock = mock.MagicMock(
            clean_finished_oob_record=mock.CoroutineMock(return_value=None)
        )
        self.route_manager = CoordinateMediationV1RouteManager()
        self.resolver = DIDResolver()
        self.resolver.register_resolver(LegacyPeerDIDResolver())
        self.resolver.register_resolver(KeyDIDResolver())
        self.resolver.register_resolver(PeerDID4Resolver())

        self.profile = await create_test_profile(
            {
                "default_endpoint": "http://aries.ca/endpoint",
                "default_label": "This guy",
                "additional_endpoints": ["http://aries.ca/another-endpoint"],
                "debug.auto_accept_invites": True,
                "debug.auto_accept_requests": True,
            },
        )
        self.profile.context.injector.bind_instance(BaseResponder, self.responder)
        self.profile.context.injector.bind_instance(BaseCache, InMemoryCache())
        self.profile.context.injector.bind_instance(OobMessageProcessor, self.oob_mock)
        self.profile.context.injector.bind_instance(RouteManager, self.route_manager)
        self.profile.context.injector.bind_instance(DIDMethods, DIDMethods())
        self.profile.context.injector.bind_instance(DIDResolver, self.resolver)
        self.profile.context.injector.bind_instance(
            KeyTypes, mock.MagicMock(KeyTypes, autospec=True)
        )
        self.context = self.profile.context

        self.multitenant_mgr = mock.MagicMock(MultitenantManager, autospec=True)
        self.context.injector.bind_instance(BaseMultitenantManager, self.multitenant_mgr)

        self.test_mediator_routing_keys = [
            "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
        ]
        self.test_mediator_conn_id = "mediator-conn-id"
        self.test_mediator_endpoint = "http://mediator.example.com"

        self.manager = BaseConnectionManager(self.profile)

        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            info = await wallet.create_local_did(method=SOV, key_type=ED25519)

        self.did = info.did
        self.verkey = info.verkey

        assert self.manager._profile

    async def test_did_key_storage(self):
        await self.manager.add_key_for_did(
            did=self.test_target_did, key=self.test_target_verkey
        )
        await self.manager.add_key_for_did(
            did=self.test_target_did, key=self.test_target_verkey
        )

        did = await self.manager.find_did_for_key(key=self.test_target_verkey)
        assert did == self.test_target_did
        await self.manager.remove_keys_for_did(self.test_target_did)

    async def test_fetch_connection_targets_no_my_did(self):
        mock_conn = mock.MagicMock()
        mock_conn.my_did = None
        assert await self.manager.fetch_connection_targets(mock_conn) == []

    async def test_fetch_connection_targets_in_progress_conn(self):
        mock_conn = mock.MagicMock(
            my_did=self.did,
            their_did=self.test_target_did,
            connection_id="dummy",
            their_role=ConnRecord.Role.RESPONDER.rfc23,
            state=ConnRecord.State.INVITATION.rfc23,
        )
        with mock.patch.object(
            self.manager,
            "_fetch_targets_for_connection_in_progress",
            mock.CoroutineMock(),
        ) as mock_fetch_in_progress:
            await self.manager.fetch_connection_targets(mock_conn)
            mock_fetch_in_progress.assert_called()

    async def test_fetch_targets_for_connection_in_progress_inv(self):
        mock_conn = mock.MagicMock(
            my_did=self.test_did,
            their_did=self.test_target_did,
            connection_id="dummy",
            their_role=ConnRecord.Role.RESPONDER.rfc23,
            state=ConnRecord.State.INVITATION.rfc23,
            invitation_msg_id="test-invite-msg-id",
        )
        mock_conn.retrieve_invitation = mock.CoroutineMock()
        with mock.patch.object(
            self.manager,
            "_fetch_connection_targets_for_invitation",
            mock.CoroutineMock(),
        ) as mock_fetch_connection_targets_for_invitation:
            await self.manager._fetch_targets_for_connection_in_progress(
                mock_conn, self.test_did
            )
            mock_fetch_connection_targets_for_invitation.assert_called()

    async def test_fetch_targets_for_connection_in_progress_implicit(self):
        mock_conn = mock.MagicMock(
            my_did=self.test_did,
            their_did=self.test_target_did,
            connection_id="dummy",
            their_role=ConnRecord.Role.RESPONDER.rfc23,
            state=ConnRecord.State.INVITATION.rfc23,
            invitation_msg_id=None,
            invitation_key=None,
        )
        with mock.patch.object(
            self.manager,
            "resolve_invitation",
            mock.CoroutineMock(
                return_value=(mock.MagicMock(), mock.MagicMock(), mock.MagicMock())
            ),
        ) as mock_resolve_invitation:
            await self.manager._fetch_targets_for_connection_in_progress(
                mock_conn, self.test_did
            )
            mock_resolve_invitation.assert_called()

    async def test_fetch_connection_targets_for_invitation_did_resolver(self):
        target_did_info = await self.manager.create_did_peer_4()
        target_did = target_did_info.did
        invitation = InvitationMessage(services=[OOBService(did=target_did)])

        mock_conn = mock.MagicMock(
            my_did=self.test_did,
            their_did=target_did,
            connection_id="dummy",
            their_role=ConnRecord.Role.RESPONDER.rfc23,
            state=ConnRecord.State.INVITATION.rfc23,
        )

        targets = await self.manager._fetch_connection_targets_for_invitation(
            mock_conn, invitation, self.test_did
        )
        assert len(targets) == 1
        target = targets[0]
        assert target.did == mock_conn.their_did

    async def test_fetch_connection_targets_conn_invitation_btcr_without_services(self):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            did_doc_json = {
                "@context": ["https://www.w3.org/ns/did/v1"],
                "id": "did:btcr:x705-jznz-q3nl-srs",
                "verificationMethod": [
                    {
                        "type": "EcdsaSecp256k1VerificationKey2019",
                        "id": "did:btcr:x705-jznz-q3nl-srs#key-0",
                        "publicKeyBase58": "02e0e01a8c302976e1556e95c54146e8464adac8626a5d29474718a7281133ff49",
                    },
                    {
                        "type": "EcdsaSecp256k1VerificationKey2019",
                        "id": "did:btcr:x705-jznz-q3nl-srs#key-1",
                        "publicKeyBase58": "02e0e01a8c302976e1556e95c54146e8464adac8626a5d29474718a7281133ff49",
                    },
                    {
                        "type": "EcdsaSecp256k1VerificationKey2019",
                        "id": "did:btcr:x705-jznz-q3nl-srs#satoshi",
                        "publicKeyBase58": "02e0e01a8c302976e1556e95c54146e8464adac8626a5d29474718a7281133ff49",
                    },
                ],
            }
            did_doc = DIDDocument.deserialize(did_doc_json)
            self.resolver.resolve = mock.CoroutineMock(return_value=did_doc)
            self.context.injector.bind_instance(DIDResolver, self.resolver)

            invitation = InvitationMessage(services=[did_doc.id])
            mock_conn = mock.MagicMock(
                my_did=did_doc.id,
                their_did=self.test_target_did,
                connection_id="dummy",
                their_role=ConnRecord.Role.RESPONDER.rfc23,
                state=ConnRecord.State.INVITATION.rfc23,
                retrieve_invitation=mock.CoroutineMock(return_value=invitation),
            )
            with self.assertRaises(BaseConnectionManagerError):
                await self.manager._fetch_connection_targets_for_invitation(
                    mock_conn, invitation, self.test_did
                )

    async def test_fetch_connection_targets_conn_invitation_no_didcomm_services(self):
        async with self.profile.session() as session:
            did_doc_json = {
                "@context": ["https://www.w3.org/ns/did/v1"],
                "id": "did:btcr:x705-jznz-q3nl-srs",
                "verificationMethod": [
                    {
                        "type": "EcdsaSecp256k1VerificationKey2019",
                        "id": "did:btcr:x705-jznz-q3nl-srs#key-0",
                        "publicKeyBase58": "02e0e01a8c302976e1556e95c54146e8464adac8626a5d29474718a7281133ff49",
                    },
                    {
                        "type": "EcdsaSecp256k1VerificationKey2019",
                        "id": "did:btcr:x705-jznz-q3nl-srs#key-1",
                        "publicKeyBase58": "02e0e01a8c302976e1556e95c54146e8464adac8626a5d29474718a7281133ff49",
                    },
                    {
                        "type": "EcdsaSecp256k1VerificationKey2019",
                        "id": "did:btcr:x705-jznz-q3nl-srs#satoshi",
                        "publicKeyBase58": "02e0e01a8c302976e1556e95c54146e8464adac8626a5d29474718a7281133ff49",
                    },
                ],
            }

            did_doc = DIDDocument.deserialize(did_doc_json)
            builder = DIDDocumentBuilder.from_doc(did_doc)

            builder.verification_method.add(
                Ed25519VerificationKey2018, public_key_base58=self.test_target_verkey
            )
            builder.service.add(type_="LinkedData", service_endpoint=self.test_endpoint)
            did_doc = builder.build()
            self.resolver.resolve = mock.CoroutineMock(return_value=did_doc)
            self.context.injector.bind_instance(DIDResolver, self.resolver)

            invitation = InvitationMessage(services=[did_doc.id])
            mock_conn = mock.MagicMock(
                my_did=did_doc.id,
                their_did=self.test_target_did,
                connection_id="dummy",
                their_role=ConnRecord.Role.RESPONDER.rfc23,
                state=ConnRecord.State.INVITATION.rfc23,
                retrieve_invitation=mock.CoroutineMock(return_value=invitation),
            )
            with self.assertRaises(BaseConnectionManagerError):
                await self.manager._fetch_connection_targets_for_invitation(
                    mock_conn, invitation, self.test_did
                )

    async def test_resolve_invitation_supports_Ed25519VerificationKey2018_key_type_no_multicodec(
        self,
    ):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            builder = DIDDocumentBuilder("did:btcr:x705-jznz-q3nl-srs")
            vmethod = builder.verification_method.add(
                Ed25519VerificationKey2020,
                public_key_multibase=multibase.encode(
                    b58_to_bytes(self.test_target_verkey), "base58btc"
                ),
            )
            builder.service.add_didcomm(
                type_="IndyAgent",
                service_endpoint=self.test_endpoint,
                recipient_keys=[vmethod],
            )
            did_doc = builder.build()
            self.resolver.resolve = mock.CoroutineMock(return_value=did_doc)
            assert did_doc.verification_method
            self.resolver.dereference = mock.CoroutineMock(
                return_value=did_doc.verification_method[0]
            )
            self.context.injector.bind_instance(DIDResolver, self.resolver)

            endpoint, recips, routes = await self.manager.resolve_invitation(did_doc.id)
            assert endpoint == self.test_endpoint
            assert recips == [self.test_target_verkey]
            assert routes == []

    async def test_resolve_invitation_supports_Ed25519VerificationKey2018_key_type_with_multicodec(
        self,
    ):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            builder = DIDDocumentBuilder("did:btcr:x705-jznz-q3nl-srs")
            vmethod = builder.verification_method.add(
                Ed25519VerificationKey2020,
                public_key_multibase=multibase.encode(
                    multicodec.wrap("ed25519-pub", b58_to_bytes(self.test_target_verkey)),
                    "base58btc",
                ),
            )
            builder.service.add_didcomm(
                type_="IndyAgent",
                service_endpoint=self.test_endpoint,
                recipient_keys=[vmethod],
            )
            did_doc = builder.build()
            self.resolver.resolve = mock.CoroutineMock(return_value=did_doc)
            assert did_doc.verification_method
            self.resolver.dereference = mock.CoroutineMock(
                return_value=did_doc.verification_method[0]
            )
            self.context.injector.bind_instance(DIDResolver, self.resolver)

            endpoint, recips, routes = await self.manager.resolve_invitation(did_doc.id)
            assert endpoint == self.test_endpoint
            assert recips == [self.test_target_verkey]
            assert routes == []

    async def test_resolve_invitation_supported_JsonWebKey2020_key_type(
        self,
    ):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            builder = DIDDocumentBuilder("did:btcr:x705-jznz-q3nl-srs")
            vmethod = builder.verification_method.add(
                JsonWebKey2020,
                ident="1",
                public_key_jwk={
                    "kty": "OKP",
                    "crv": "Ed25519",
                    "x": bytes_to_b64(b58_to_bytes(self.test_target_verkey), True),
                },
            )
            builder.service.add_didcomm(
                type_="IndyAgent",
                service_endpoint=self.test_endpoint,
                recipient_keys=[vmethod],
            )
            did_doc = builder.build()
            self.resolver.resolve = mock.CoroutineMock(return_value=did_doc)
            assert did_doc.verification_method
            self.resolver.dereference = mock.CoroutineMock(
                return_value=did_doc.verification_method[0]
            )
            self.context.injector.bind_instance(DIDResolver, self.resolver)

            endpoint, recips, routes = await self.manager.resolve_invitation(did_doc.id)
            assert endpoint == self.test_endpoint
            assert recips == [self.test_target_verkey]
            assert routes == []

    async def test_resolve_invitation_unsupported_key_type(self):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            builder = DIDDocumentBuilder("did:btcr:x705-jznz-q3nl-srs")
            vmethod = builder.verification_method.add(
                JsonWebKey2020,
                ident="1",
                public_key_jwk={
                    "kty": "EC",
                    "crv": "P-256",
                    "x": "2syLh57B-dGpa0F8p1JrO6JU7UUSF6j7qL-vfk1eOoY",
                    "y": "BgsGtI7UPsObMRjdElxLOrgAO9JggNMjOcfzEPox18w",
                },
            )
            builder.service.add_didcomm(
                type_="IndyAgent",
                service_endpoint=self.test_endpoint,
                recipient_keys=[vmethod],
            )
            did_doc = builder.build()
            self.resolver.resolve = mock.CoroutineMock(return_value=did_doc)
            assert did_doc.verification_method
            self.resolver.dereference = mock.CoroutineMock(
                return_value=did_doc.verification_method[0]
            )
            self.context.injector.bind_instance(DIDResolver, self.resolver)

            with self.assertRaises(BaseConnectionManagerError):
                await self.manager.resolve_invitation(did_doc.id)

    async def test_fetch_connection_targets_conn_initiator_completed_no_their_did(self):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
                seed=self.test_seed,
                did=self.test_did,
                metadata=None,
            )

            mock_conn = mock.MagicMock(
                my_did=self.test_did,
                their_did=None,
                state=ConnRecord.State.COMPLETED.rfc23,
            )
            assert await self.manager.fetch_connection_targets(mock_conn) == []

    async def test_fetch_connection_targets_conn_completed_their_did(self):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            local_did = await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
                seed=self.test_seed,
                did=self.test_did,
                metadata=None,
            )

            did_doc = self.make_did_doc(did=self.test_did, verkey=self.test_verkey)
            await self.manager.store_did_document(did_doc)

            mock_conn = mock.MagicMock(
                my_did=self.test_did,
                their_did=self.test_did,
                their_label="label",
                their_role=ConnRecord.Role.REQUESTER.rfc160,
                state=ConnRecord.State.COMPLETED.rfc23,
            )

            targets = await self.manager.fetch_connection_targets(mock_conn)
            assert len(targets) == 1
            target = targets[0]
            # did:sov: dropped for this check
            assert target.did[8:] == mock_conn.their_did
            assert target.endpoint == self.test_endpoint
            assert target.label == mock_conn.their_label
            assert target.recipient_keys == [self.test_verkey]
            assert target.routing_keys == []
            assert target.sender_key == local_did.verkey

    async def test_fetch_connection_targets_conn_no_invi_with_their_did(self):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            local_did = await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
                seed=self.test_seed,
                did=self.test_did,
                metadata=None,
            )

            self.manager.resolve_invitation = mock.CoroutineMock()
            self.manager.resolve_invitation.return_value = (
                self.test_endpoint,
                [self.test_verkey],
                [],
            )

            did_doc = self.make_did_doc(did=self.test_did, verkey=self.test_verkey)
            await self.manager.store_did_document(did_doc)

            mock_conn = mock.MagicMock(
                my_did=self.test_did,
                their_did=self.test_did,
                their_label="label",
                their_role=ConnRecord.Role.RESPONDER.rfc23,
                state=ConnRecord.State.REQUEST.rfc23,
                invitation_key=None,
                invitation_msg_id=None,
            )

            targets = await self.manager.fetch_connection_targets(mock_conn)
            assert len(targets) == 1
            target = targets[0]
            assert target.did == mock_conn.their_did
            assert target.endpoint == self.test_endpoint
            assert target.label is None
            assert target.recipient_keys == [self.test_verkey]
            assert target.routing_keys == []
            assert target.sender_key == local_did.verkey

    async def test_verification_methods_for_service(self):
        did = "did:sov:" + self.test_did
        doc_builder = DIDDocumentBuilder(did)
        vm = doc_builder.verification_method.add(
            Ed25519VerificationKey2018,
            public_key_base58=self.test_verkey,
        )
        route_key = DIDKey.from_public_key_b58(self.test_verkey, ED25519)
        service = doc_builder.service.add(
            type_="did-communication",
            service_endpoint=self.test_endpoint,
            recipient_keys=[vm.id],
            routing_keys=[route_key.key_id],
        )
        doc = doc_builder.build()
        self.manager.resolve_didcomm_services = mock.CoroutineMock(
            return_value=(doc, doc.service)
        )
        recip, routing = await self.manager.verification_methods_for_service(doc, service)
        assert recip == [vm]
        assert routing

    async def test_resolve_connection_targets_empty(self):
        """Test resolve connection targets."""
        did = "did:sov:" + self.test_did
        self.manager.resolve_didcomm_services = mock.CoroutineMock(
            return_value=(DIDDocument(id=DID(did)), [])
        )
        targets = await self.manager.resolve_connection_targets(did)
        assert targets == []

    async def test_resolve_connection_targets(self):
        """Test resolve connection targets."""
        did = "did:sov:" + self.test_did
        doc_builder = DIDDocumentBuilder(did)
        vm = doc_builder.verification_method.add(
            Ed25519VerificationKey2018,
            public_key_base58=self.test_verkey,
        )
        route_key = DIDKey.from_public_key_b58(self.test_verkey, ED25519)
        doc_builder.service.add(
            type_="did-communication",
            service_endpoint=self.test_endpoint,
            recipient_keys=[vm.id],
            routing_keys=[route_key.key_id],
        )
        doc = doc_builder.build()
        self.manager.resolve_didcomm_services = mock.CoroutineMock(
            return_value=(doc, doc.service)
        )
        targets = await self.manager.resolve_connection_targets(did)
        assert targets
        assert targets[0].routing_keys[0] == self.test_verkey

    async def test_resolve_connection_targets_x_bad_reference(self):
        """Test resolve connection targets."""
        did = "did:sov:" + self.test_did
        doc_builder = DIDDocumentBuilder(did)
        vm = doc_builder.verification_method.add(
            Ed25519VerificationKey2018,
            public_key_base58=self.test_verkey,
        )
        doc_builder.service.add(
            type_="did-communication",
            service_endpoint=self.test_endpoint,
            recipient_keys=[vm.id],
            routing_keys=["did:example:123#some-random-id"],
        )
        doc = doc_builder.build()
        self.manager.resolve_didcomm_services = mock.CoroutineMock(
            return_value=(doc, doc.service)
        )
        with self.assertLogs() as cm:
            await self.manager.resolve_connection_targets(did)
        assert cm.output and "Failed to resolve service" in cm.output[0]

    async def test_resolve_connection_targets_x_bad_key_material(self):
        did = "did:sov:" + self.test_did
        doc_builder = DIDDocumentBuilder(did)
        vm = doc_builder.verification_method.add(
            Ed25519VerificationKey2020,
            public_key_multibase=multibase.encode(
                multicodec.wrap("secp256k1-pub", b58_to_bytes(self.test_verkey)),
                "base58btc",
            ),
        )
        route_key = DIDKey.from_public_key_b58(self.test_verkey, ED25519)
        doc_builder.service.add(
            type_="did-communication",
            service_endpoint=self.test_endpoint,
            recipient_keys=[vm.id],
            routing_keys=[route_key.key_id],
        )
        doc = doc_builder.build()
        self.manager.resolve_didcomm_services = mock.CoroutineMock(
            return_value=(doc, doc.service)
        )
        with self.assertRaises(BaseConnectionManagerError) as cm:
            await self.manager.resolve_connection_targets(did)
        assert "not supported" in str(cm.exception)

    @pytest.mark.filterwarnings("ignore::UserWarning")
    async def test_resolve_connection_targets_x_unsupported_key(self):
        did = "did:sov:" + self.test_did
        doc_builder = DIDDocumentBuilder(did)
        vm = doc_builder.verification_method.add(
            EcdsaSecp256k1VerificationKey2019,
            public_key_hex="deadbeef",
        )
        route_key = DIDKey.from_public_key_b58(self.test_verkey, ED25519)
        doc_builder.service.add(
            type_="did-communication",
            service_endpoint=self.test_endpoint,
            recipient_keys=[vm.id],
            routing_keys=[route_key.key_id],
        )
        doc = doc_builder.build()
        self.manager.resolve_didcomm_services = mock.CoroutineMock(
            return_value=(doc, doc.service)
        )
        with self.assertRaises(BaseConnectionManagerError) as cm:
            await self.manager.resolve_connection_targets(did)
        assert "not supported" in str(cm.exception)

    async def test_record_keys_for_resolvable_did_empty(self):
        did = "did:sov:" + self.test_did
        service_builder = ServiceBuilder(DID(did))
        service_builder.add_didcomm(
            self.test_endpoint, recipient_keys=[], routing_keys=[]
        )
        self.manager.resolve_didcomm_services = mock.CoroutineMock(
            return_value=(DIDDocument(id=DID(did)), service_builder.services)
        )
        await self.manager.record_keys_for_resolvable_did(did)

    async def test_record_keys_for_resolvable_did(self):
        did = "did:sov:" + self.test_did
        doc_builder = DIDDocumentBuilder(did)
        vm = doc_builder.verification_method.add(
            Ed25519VerificationKey2018,
            public_key_base58=self.test_verkey,
        )
        doc_builder.service.add_didcomm(
            self.test_endpoint, recipient_keys=[vm], routing_keys=[]
        )
        doc = doc_builder.build()
        self.manager.resolve_didcomm_services = mock.CoroutineMock(
            return_value=(doc, doc.service)
        )
        await self.manager.record_keys_for_resolvable_did(did)

    async def test_diddoc_connection_targets_diddoc_underspecified(self):
        with self.assertRaises(BaseConnectionManagerError):
            self.manager.diddoc_connection_targets(None, self.test_verkey)

        x_did_doc = DIDDoc(did=None)
        with self.assertRaises(BaseConnectionManagerError):
            self.manager.diddoc_connection_targets(x_did_doc, self.test_verkey)

        x_did_doc = self.make_did_doc(
            did=self.test_target_did, verkey=self.test_target_verkey
        )
        x_did_doc._service = {}
        with self.assertRaises(BaseConnectionManagerError):
            self.manager.diddoc_connection_targets(x_did_doc, self.test_verkey)

    async def test_find_inbound_connection(self):
        receipt = MessageReceipt(
            sender_verkey=self.test_verkey,
            recipient_verkey=self.test_target_verkey,
            recipient_did_public=False,
        )

        mock_conn = mock.MagicMock()
        mock_conn.connection_id = "dummy"

        # First pass: not yet in cache
        with mock.patch.object(
            BaseConnectionManager,
            "resolve_inbound_connection",
            mock.CoroutineMock(),
        ) as mock_conn_mgr_resolve_conn:
            mock_conn_mgr_resolve_conn.return_value = mock_conn

            conn_rec = await self.manager.find_inbound_connection(receipt)
            assert conn_rec

        # Second pass: in cache
        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve_by_id:
            mock_conn_rec_retrieve_by_id.return_value = mock_conn

            conn_rec = await self.manager.find_inbound_connection(receipt)
            assert conn_rec.id == mock_conn.id

    async def test_find_inbound_connection_no_cache(self):
        receipt = MessageReceipt(
            sender_verkey=self.test_verkey,
            recipient_verkey=self.test_target_verkey,
            recipient_did_public=False,
        )

        mock_conn = mock.MagicMock()
        mock_conn.connection_id = "dummy"

        with mock.patch.object(
            BaseConnectionManager,
            "resolve_inbound_connection",
            mock.CoroutineMock(),
        ) as mock_conn_mgr_resolve_conn:
            self.context.injector.clear_binding(BaseCache)
            mock_conn_mgr_resolve_conn.return_value = mock_conn

            conn_rec = await self.manager.find_inbound_connection(receipt)
            assert conn_rec

    async def test_resolve_inbound_connection(self):
        receipt = MessageReceipt(
            sender_verkey=self.test_verkey,
            recipient_verkey=self.test_target_verkey,
            recipient_did_public=True,
        )

        mock_conn = mock.MagicMock()
        mock_conn.connection_id = "dummy"

        with (
            mock.patch.object(
                AskarWallet, "get_local_did_for_verkey", mock.CoroutineMock()
            ) as mock_wallet_get_local_did_for_verkey,
            mock.patch.object(
                self.manager, "find_connection", mock.CoroutineMock()
            ) as mock_mgr_find_conn,
        ):
            mock_wallet_get_local_did_for_verkey.return_value = DIDInfo(
                self.test_did,
                self.test_verkey,
                {"posted": True},
                method=SOV,
                key_type=ED25519,
            )
            mock_mgr_find_conn.return_value = mock_conn

            assert await self.manager.resolve_inbound_connection(receipt)

    async def test_resolve_inbound_connection_injector_error(self):
        receipt = MessageReceipt(
            sender_verkey=self.test_verkey,
            recipient_verkey=self.test_target_verkey,
            recipient_did_public=True,
        )

        mock_conn = mock.MagicMock()
        mock_conn.connection_id = "dummy"

        with (
            mock.patch.object(
                AskarWallet, "get_local_did_for_verkey", mock.CoroutineMock()
            ) as mock_wallet_get_local_did_for_verkey,
            mock.patch.object(
                self.manager, "find_connection", mock.CoroutineMock()
            ) as mock_mgr_find_conn,
        ):
            mock_wallet_get_local_did_for_verkey.side_effect = InjectionError()
            mock_mgr_find_conn.return_value = mock_conn

            assert await self.manager.resolve_inbound_connection(receipt)

    async def test_resolve_inbound_connection_wallet_not_found_error(self):
        receipt = MessageReceipt(
            sender_verkey=self.test_verkey,
            recipient_verkey=self.test_target_verkey,
            recipient_did_public=True,
        )

        mock_conn = mock.MagicMock()
        mock_conn.connection_id = "dummy"

        with (
            mock.patch.object(
                AskarWallet, "get_local_did_for_verkey", mock.CoroutineMock()
            ) as mock_wallet_get_local_did_for_verkey,
            mock.patch.object(
                self.manager, "find_connection", mock.CoroutineMock()
            ) as mock_mgr_find_conn,
        ):
            mock_wallet_get_local_did_for_verkey.side_effect = WalletNotFoundError()
            mock_mgr_find_conn.return_value = mock_conn

            assert await self.manager.resolve_inbound_connection(receipt)

    async def test_get_connection_targets_retrieve_connection(self):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            local_did = await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
                seed=self.test_seed,
                did=self.test_did,
                metadata=None,
            )

            did_doc = self.make_did_doc(
                did=self.test_target_did, verkey=self.test_target_verkey
            )
            await self.manager.store_did_document(did_doc)

            # Connection target not in cache
            service = OOBService(
                did=None,
                recipient_keys=[
                    DIDKey.from_public_key_b58(self.test_target_verkey, ED25519).did
                ],
                routing_keys=[DIDKey.from_public_key_b58(self.test_verkey, ED25519).did],
            )
            invitation = InvitationMessage(
                services=[service],
                label="label",
            )
            mock_conn = mock.MagicMock(
                my_did=self.test_did,
                their_did=self.test_target_did,
                connection_id="dummy",
                their_role=ConnRecord.Role.RESPONDER.rfc23,
                state=ConnRecord.State.INVITATION.rfc23,
                retrieve_invitation=mock.CoroutineMock(return_value=invitation),
            )

            with (
                mock.patch.object(
                    ConnectionTarget, "serialize", autospec=True
                ) as mock_conn_target_ser,
                mock.patch.object(
                    ConnRecord, "retrieve_by_id", mock.CoroutineMock()
                ) as mock_conn_rec_retrieve_by_id,
            ):
                mock_conn_rec_retrieve_by_id.return_value = mock_conn
                mock_conn_target_ser.return_value = {"serialized": "value"}
                targets = await self.manager.get_connection_targets(
                    connection_id="dummy",
                    connection=None,
                )
                assert len(targets) == 1
                target = targets[0]
                assert target.did == mock_conn.their_did
                assert target.endpoint == service.service_endpoint
                assert target.label == invitation.label
                assert target.recipient_keys == [self.test_target_verkey]
                assert target.routing_keys == [self.test_verkey]
                assert target.sender_key == local_did.verkey

    async def test_get_connection_targets_from_cache(self):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
                seed=self.test_seed,
                did=self.test_did,
                metadata=None,
            )

            did_doc = self.make_did_doc(
                did=self.test_target_did, verkey=self.test_target_verkey
            )
            await self.manager.store_did_document(did_doc)

            mock_conn = mock.MagicMock(
                my_did=self.test_did,
                their_did=self.test_target_did,
                connection_id="dummy",
                their_role=ConnRecord.Role.RESPONDER.rfc23,
                state=ConnRecord.State.COMPLETED.rfc160,
            )

            with (
                mock.patch.object(
                    ConnectionTarget, "serialize", autospec=True
                ) as mock_conn_target_ser,
                mock.patch.object(
                    ConnRecord, "retrieve_by_id", mock.CoroutineMock()
                ) as mock_conn_rec_retrieve_by_id,
                mock.patch.object(
                    self.manager, "fetch_connection_targets", mock.CoroutineMock()
                ) as mock_fetch_connection_targets,
            ):
                mock_fetch_connection_targets.return_value = [ConnectionTarget()]
                mock_conn_rec_retrieve_by_id.return_value = mock_conn
                mock_conn_target_ser.return_value = {"serialized": "value"}
                await self.manager.get_connection_targets(
                    connection_id="dummy",
                    connection=None,
                )

                await self.manager.get_connection_targets(
                    connection_id="dummy",
                    connection=None,
                )
                assert mock_fetch_connection_targets.call_count == 1

    async def test_get_connection_targets_no_cache(self):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
                seed=self.test_seed,
                did=self.test_did,
                metadata=None,
            )

            did_doc = self.make_did_doc(
                did=self.test_target_did, verkey=self.test_target_verkey
            )
            await self.manager.store_did_document(did_doc)

            mock_conn = mock.MagicMock(
                my_did=self.test_did,
                their_did=self.test_target_did,
                connection_id="dummy",
                their_role=ConnRecord.Role.RESPONDER.rfc23,
                state=ConnRecord.State.COMPLETED.rfc160,
            )

            with (
                mock.patch.object(
                    ConnectionTarget, "serialize", autospec=True
                ) as mock_conn_target_ser,
                mock.patch.object(
                    ConnRecord, "retrieve_by_id", mock.CoroutineMock()
                ) as mock_conn_rec_retrieve_by_id,
                mock.patch.object(
                    self.manager, "fetch_connection_targets", mock.CoroutineMock()
                ) as mock_fetch_connection_targets,
            ):
                mock_fetch_connection_targets.return_value = [ConnectionTarget()]
                mock_conn_rec_retrieve_by_id.return_value = mock_conn
                mock_conn_target_ser.return_value = {"serialized": "value"}
                self.profile.context.injector.clear_binding(BaseCache)
                targets = await self.manager.get_connection_targets(
                    connection_id="dummy",
                    connection=None,
                )
                assert targets
                targets = await self.manager.get_connection_targets(
                    connection_id=None,
                    connection=mock_conn,
                )
                assert targets

    async def test_get_connection_targets_no_conn_or_id(self):
        with self.assertRaises(ValueError):
            await self.manager.get_connection_targets()

    async def test_get_conn_targets_invitation_no_cache(self):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            self.context.injector.clear_binding(BaseCache)
            local_did = await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
                seed=self.test_seed,
                did=self.test_did,
                metadata=None,
            )

            did_doc = self.make_did_doc(
                did=self.test_target_did, verkey=self.test_target_verkey
            )
            await self.manager.store_did_document(did_doc)

            service = OOBService(
                did=None,
                recipient_keys=[
                    DIDKey.from_public_key_b58(self.test_target_verkey, ED25519).did
                ],
                routing_keys=[DIDKey.from_public_key_b58(self.test_verkey, ED25519).did],
            )
            invitation = InvitationMessage(
                services=[service],
                label="label",
            )
            mock_conn = mock.MagicMock(
                my_did=self.test_did,
                their_did=self.test_target_did,
                connection_id="dummy",
                their_role=ConnRecord.Role.RESPONDER.rfc23,
                state=ConnRecord.State.INVITATION.rfc23,
                retrieve_invitation=mock.CoroutineMock(return_value=invitation),
            )

            targets = await self.manager.get_connection_targets(
                connection_id=None,
                connection=mock_conn,
            )
            assert len(targets) == 1
            target = targets[0]
            assert target.did == mock_conn.their_did
            assert target.endpoint == service.service_endpoint
            assert target.label == invitation.label
            assert target.recipient_keys == [self.test_target_verkey]
            assert target.routing_keys == [self.test_verkey]
            assert target.sender_key == local_did.verkey

    @pytest.mark.filterwarnings("ignore::UserWarning")  # create_did_document deprecation
    async def test_create_static_connection(self):
        self.multitenant_mgr = mock.MagicMock(MultitenantManager, autospec=True)
        self.multitenant_mgr.get_default_mediator = mock.CoroutineMock(return_value=None)
        self.context.injector.bind_instance(BaseMultitenantManager, self.multitenant_mgr)

        _my, _their, conn_rec = await self.manager.create_static_connection(
            my_did=base58.b58encode(secrets.token_bytes(16)).decode(),
            their_did=self.test_target_did,
            their_verkey=self.test_target_verkey,
            their_endpoint=self.test_endpoint,
        )

        assert ConnRecord.State.get(conn_rec.state) is ConnRecord.State.COMPLETED

    @pytest.mark.filterwarnings("ignore::UserWarning")  # create_did_document deprecation
    async def test_create_static_connection_multitenant(self):
        self.context.update_settings(
            {"wallet.id": "test_wallet", "multitenant.enabled": True}
        )

        self.multitenant_mgr.get_default_mediator.return_value = None
        self.route_manager.route_static = mock.CoroutineMock()

        with (
            mock.patch.object(ConnRecord, "save", autospec=True),
            mock.patch.object(
                AskarWallet, "create_local_did", autospec=True
            ) as mock_wallet_create_local_did,
        ):
            mock_wallet_create_local_did.return_value = DIDInfo(
                self.test_did,
                self.test_verkey,
                None,
                method=SOV,
                key_type=ED25519,
            )

            await self.manager.create_static_connection(
                my_did=self.test_did,
                their_did=self.test_target_did,
                their_verkey=self.test_target_verkey,
                their_endpoint=self.test_endpoint,
            )

            self.route_manager.route_static.assert_called_once()

    @pytest.mark.filterwarnings("ignore::UserWarning")  # create_did_document deprecation
    async def test_create_static_connection_multitenant_auto_disclose_features(self):
        self.context.update_settings(
            {
                "auto_disclose_features": True,
                "multitenant.enabled": True,
                "wallet.id": "test_wallet",
            }
        )
        self.multitenant_mgr.get_default_mediator.return_value = None
        self.route_manager.route_static = mock.CoroutineMock()
        with (
            mock.patch.object(ConnRecord, "save", autospec=True),
            mock.patch.object(
                AskarWallet, "create_local_did", autospec=True
            ) as mock_wallet_create_local_did,
            mock.patch.object(
                V20DiscoveryMgr, "proactive_disclose_features", mock.CoroutineMock()
            ) as mock_proactive_disclose_features,
        ):
            mock_wallet_create_local_did.return_value = DIDInfo(
                self.test_did,
                self.test_verkey,
                None,
                method=SOV,
                key_type=ED25519,
            )
            await self.manager.create_static_connection(
                my_did=self.test_did,
                their_did=self.test_target_did,
                their_verkey=self.test_target_verkey,
                their_endpoint=self.test_endpoint,
            )
            self.route_manager.route_static.assert_called_once()
            mock_proactive_disclose_features.assert_called_once()

    async def test_create_static_connection_multitenant_mediator(self):
        self.context.update_settings(
            {"wallet.id": "test_wallet", "multitenant.enabled": True}
        )

        default_mediator = mock.MagicMock()
        self.route_manager.route_static = mock.CoroutineMock()

        with (
            mock.patch.object(ConnRecord, "save", autospec=True),
            mock.patch.object(
                AskarWallet, "create_local_did", autospec=True
            ) as mock_wallet_create_local_did,
            mock.patch.object(
                BaseConnectionManager, "create_did_document"
            ) as create_did_document,
            mock.patch.object(BaseConnectionManager, "store_did_document"),
        ):
            mock_wallet_create_local_did.return_value = DIDInfo(
                self.test_did,
                self.test_verkey,
                None,
                method=SOV,
                key_type=ED25519,
            )

            # With default mediator
            self.multitenant_mgr.get_default_mediator.return_value = default_mediator
            await self.manager.create_static_connection(
                my_did=self.test_did,
                their_did=self.test_target_did,
                their_verkey=self.test_target_verkey,
                their_endpoint=self.test_endpoint,
            )

            # Without default mediator
            self.multitenant_mgr.get_default_mediator.return_value = None
            await self.manager.create_static_connection(
                my_did=self.test_did,
                their_did=self.test_target_did,
                their_verkey=self.test_target_verkey,
                their_endpoint=self.test_endpoint,
            )

            assert self.route_manager.route_static.call_count == 2

            their_info = DIDInfo(
                self.test_target_did,
                self.test_target_verkey,
                {},
                method=SOV,
                key_type=ED25519,
            )
            create_did_document.assert_has_calls(
                [
                    call(
                        their_info,
                        [self.test_endpoint],
                        mediation_records=[default_mediator],
                    ),
                    call(their_info, [self.test_endpoint], mediation_records=[]),
                ]
            )

    async def test_create_static_connection_no_their(self):
        with mock.patch.object(ConnRecord, "save", autospec=True):
            with self.assertRaises(BaseConnectionManagerError):
                await self.manager.create_static_connection(
                    my_did=self.test_did,
                    their_did=None,
                    their_verkey=self.test_target_verkey,
                    their_endpoint=self.test_endpoint,
                )

    @pytest.mark.filterwarnings("ignore::UserWarning")  # create_did_document deprecation
    async def test_create_static_connection_their_seed_only(self):
        self.multitenant_mgr = mock.MagicMock(MultitenantManager, autospec=True)
        self.multitenant_mgr.get_default_mediator = mock.CoroutineMock(return_value=None)
        self.context.injector.bind_instance(BaseMultitenantManager, self.multitenant_mgr)
        _my, _their, conn_rec = await self.manager.create_static_connection(
            my_did=self.test_did,
            their_seed=self.test_seed,
            their_endpoint=self.test_endpoint,
        )

        assert ConnRecord.State.get(conn_rec.state) is ConnRecord.State.COMPLETED

    async def test_find_connection_retrieve_by_did(self):
        with mock.patch.object(
            ConnRecord, "retrieve_by_did", mock.CoroutineMock()
        ) as mock_conn_retrieve_by_did:
            mock_conn_retrieve_by_did.return_value = mock.MagicMock(
                state=ConnRecord.State.RESPONSE.rfc23,
                save=mock.CoroutineMock(),
            )

            conn_rec = await self.manager.find_connection(
                their_did=self.test_target_did,
                my_did=self.test_did,
                parent_thread_id=self.test_pthid,
                auto_complete=True,
            )
            assert ConnRecord.State.get(conn_rec.state) is ConnRecord.State.COMPLETED

    async def test_find_connection_retrieve_by_did_auto_disclose_features(self):
        self.context.update_settings({"auto_disclose_features": True})
        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_did", mock.CoroutineMock()
            ) as mock_conn_retrieve_by_did,
            mock.patch.object(
                V20DiscoveryMgr, "proactive_disclose_features", mock.CoroutineMock()
            ) as mock_proactive_disclose_features,
        ):
            mock_conn_retrieve_by_did.return_value = mock.MagicMock(
                state=ConnRecord.State.RESPONSE.rfc23,
                save=mock.CoroutineMock(),
            )

            conn_rec = await self.manager.find_connection(
                their_did=self.test_target_did,
                my_did=self.test_did,
                parent_thread_id=self.test_pthid,
                auto_complete=True,
            )
            assert ConnRecord.State.get(conn_rec.state) is ConnRecord.State.COMPLETED
            mock_proactive_disclose_features.assert_called_once()

    async def test_find_connection_retrieve_by_invitation_key(self):
        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_did", mock.CoroutineMock()
            ) as mock_conn_retrieve_by_did,
            mock.patch.object(
                ConnRecord, "retrieve_by_invitation_msg_id", mock.CoroutineMock()
            ) as mock_conn_retrieve_by_invitation_msg_id,
        ):
            mock_conn_retrieve_by_did.side_effect = StorageNotFoundError()
            mock_conn_retrieve_by_invitation_msg_id.return_value = mock.MagicMock(
                state=ConnRecord.State.RESPONSE,
                save=mock.CoroutineMock(),
            )

            conn_rec = await self.manager.find_connection(
                their_did=self.test_target_did,
                my_did=self.test_did,
                parent_thread_id=self.test_pthid,
            )
            assert conn_rec

    async def test_find_connection_retrieve_none_by_invitation_key(self):
        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_did", mock.CoroutineMock()
            ) as mock_conn_retrieve_by_did,
            mock.patch.object(
                ConnRecord, "retrieve_by_invitation_key", mock.CoroutineMock()
            ) as mock_conn_retrieve_by_invitation_msg_id,
        ):
            mock_conn_retrieve_by_did.side_effect = StorageNotFoundError()
            mock_conn_retrieve_by_invitation_msg_id.return_value = None

            conn_rec = await self.manager.find_connection(
                their_did=self.test_target_did,
                my_did=self.test_did,
                parent_thread_id=self.test_pthid,
            )
            assert conn_rec is None

    async def test_get_endpoints(self):
        conn_id = "dummy"

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_retrieve,
            mock.patch.object(
                AskarWallet, "get_local_did", autospec=True
            ) as mock_wallet_get_local_did,
            mock.patch.object(
                self.manager, "get_connection_targets", mock.CoroutineMock()
            ) as mock_get_conn_targets,
        ):
            mock_retrieve.return_value = mock.MagicMock()
            mock_wallet_get_local_did.return_value = mock.MagicMock(
                metadata={"endpoint": "localhost:8020"}
            )
            mock_get_conn_targets.return_value = [
                mock.MagicMock(endpoint="10.20.30.40:5060")
            ]
            assert await self.manager.get_endpoints(conn_id) == (
                "localhost:8020",
                "10.20.30.40:5060",
            )

    async def test_diddoc_connection_targets_diddoc(self):
        did_doc = self.make_did_doc(
            self.test_target_did,
            self.test_target_verkey,
        )
        targets = self.manager.diddoc_connection_targets(
            did_doc,
            self.test_verkey,
        )
        assert isinstance(targets[0], ConnectionTarget)



import pytest
import pytest_asyncio

from acapy_agent.tests import mock

from ......messaging.base_handler import HandlerException
from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......utils.testing import create_test_profile
from ...manager import DIDXManagerError
from ...messages.problem_report import DIDXProblemReport
from .. import problem_report_handler as test_module


@pytest_asyncio.fixture
async def request_context():
    yield RequestContext.test_context(await create_test_profile())


class TestDIDXProblemReportHandler:
    """Unit test problem report handler."""

    @pytest.mark.asyncio
    @mock.patch.object(test_module, "DIDXManager")
    async def test_called(self, manager, request_context):
        manager.return_value.receive_problem_report = mock.CoroutineMock()
        request_context.message = DIDXProblemReport()
        request_context.connection_record = mock.MagicMock()
        handler_inst = test_module.DIDXProblemReportHandler()
        responder = MockResponder()
        await handler_inst.handle(request_context, responder)
        assert not responder.messages
        manager.return_value.receive_problem_report.assert_called_once()

    @pytest.mark.asyncio
    @mock.patch.object(test_module, "DIDXManager")
    async def test_called_no_conn(self, manager, request_context):
        manager.return_value.receive_problem_report = mock.CoroutineMock()
        request_context.message = DIDXProblemReport()
        handler_inst = test_module.DIDXProblemReportHandler()
        responder = MockResponder()
        with pytest.raises(HandlerException):
            await handler_inst.handle(request_context, responder)

    @pytest.mark.asyncio
    @mock.patch.object(test_module, "DIDXManager")
    async def test_called_unrecognized_report_exception(
        self, manager, request_context, caplog
    ):
        manager.return_value.receive_problem_report = mock.CoroutineMock(
            side_effect=DIDXManagerError()
        )
        request_context.message = DIDXProblemReport()
        request_context.connection_record = mock.MagicMock()
        handler_inst = test_module.DIDXProblemReportHandler()
        responder = MockResponder()
        await handler_inst.handle(request_context, responder)
        assert "Error receiving DID Exchange problem report" in caplog.text



from .issue import issue as issue_vc
from .models import _CredentialSchema as CredentialSchema
from .models import _LDProof as LDProof
from .models import _LinkedDataProofSchema as LinkedDataProofSchema
from .models import _VerifiableCredential as VerifiableCredential
from .models import _VerifiableCredentialSchema as VerifiableCredentialSchema
from .prove import create_presentation, derive_credential, sign_presentation
from .validation_result import PresentationVerificationResult
from .verify import verify_credential, verify_presentation

__all__ = [
    "issue_vc",
    "verify_presentation",
    "verify_credential",
    "create_presentation",
    "sign_presentation",
    "derive_credential",
    "PresentationVerificationResult",
    "VerifiableCredential",
    "LDProof",
    "LinkedDataProofSchema",
    "CredentialSchema",
    "VerifiableCredentialSchema",
]



"""DID manager for Indy."""

from aries_askar import AskarError, Key

from ...core.profile import Profile
from ...utils.general import strip_did_prefix
from ...wallet.askar import CATEGORY_DID
from ...wallet.crypto import validate_seed
from ...wallet.did_method import INDY, DIDMethods
from ...wallet.did_parameters_validation import DIDParametersValidation
from ...wallet.error import WalletError
from ...wallet.key_type import ED25519, KeyType, KeyTypes
from ...wallet.util import bytes_to_b58


class DidIndyManager:
    """DID manager for Indy."""

    def __init__(self, profile: Profile) -> None:
        """Initialize the DID  manager."""
        self.profile = profile

    async def _get_holder_defined_did(self, options: dict) -> str | None:
        async with self.profile.session() as session:
            did_methods = session.inject(DIDMethods)
            indy_method = did_methods.from_method(INDY.method_name)
            did = options.get("did")

            if indy_method.holder_defined_did() and did:
                return strip_did_prefix(did)

        return None

    async def _get_key_type(self, key_type: str, default: KeyType = ED25519) -> KeyType:
        async with self.profile.session() as session:
            key_types = session.inject(KeyTypes)
            return key_types.from_key_type(key_type) or default

    def _create_key_pair(self, options: dict, key_type: KeyType) -> Key:
        seed = options.get("seed")
        if seed:
            seed = validate_seed(seed)
            return Key.from_secret_bytes(key_type, seed)
        return Key.generate(key_type)

    async def register(self, options: dict) -> dict:
        """Register a DID Indy."""
        options = options or {}
        key_type = options.get("key_type", "")

        key_type = await self._get_key_type(key_type)
        did_validation = DIDParametersValidation(self.profile.inject(DIDMethods))
        did_validation.validate_key_type(INDY, key_type)

        key_pair = self._create_key_pair(options, key_type.key_type)
        verkey_bytes = key_pair.get_public_bytes()
        verkey = bytes_to_b58(verkey_bytes)

        nym = did_validation.validate_or_derive_did(
            INDY, ED25519, verkey_bytes, (await self._get_holder_defined_did(options))
        )
        did = f"did:indy:{nym}"

        async with self.profile.session() as session:
            try:
                await session.handle.insert_key(verkey, key_pair)
                await session.handle.insert(
                    CATEGORY_DID,
                    did,
                    value_json={
                        "did": did,
                        "method": INDY.method_name,
                        "verkey": verkey,
                        "verkey_type": ED25519.key_type,
                        "metadata": {},
                    },
                    tags={
                        "method": INDY.method_name,
                        "verkey": verkey,
                        "verkey_type": ED25519.key_type,
                    },
                )
            except AskarError as err:
                raise WalletError(f"Error registering DID: {err}") from err

        return {
            "did": did,
            "verkey": verkey,
        }



from unittest import IsolatedAsyncioTestCase

from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......tests import mock
from ......transport.inbound.receipt import MessageReceipt
from ......utils.testing import create_test_profile
from ...messages.cred_problem_report import ProblemReportReason, V20CredProblemReport
from .. import cred_problem_report_handler as test_module


class TestCredProblemReportHandler(IsolatedAsyncioTestCase):
    async def test_called(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.connection_record = mock.MagicMock()

        with mock.patch.object(
            test_module, "V20CredManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value.receive_problem_report = mock.CoroutineMock()
            request_context.connection_ready = True
            request_context.message = V20CredProblemReport(
                description={
                    "en": "oh no",
                    "code": ProblemReportReason.ISSUANCE_ABANDONED.value,
                }
            )
            handler = test_module.CredProblemReportHandler()
            responder = MockResponder()
            await handler.handle(request_context, responder)

        mock_cred_mgr.assert_called_once_with(request_context.profile)
        mock_cred_mgr.return_value.receive_problem_report.assert_called_once_with(
            request_context.message, request_context.connection_record.connection_id
        )
        assert not responder.messages

    async def test_called_x(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.connection_record = mock.MagicMock()

        with mock.patch.object(
            test_module, "V20CredManager", autospec=True
        ) as mock_cred_mgr:
            request_context.connection_ready = True
            mock_cred_mgr.return_value.receive_problem_report = mock.CoroutineMock(
                side_effect=test_module.StorageError("Disk full")
            )
            request_context.message = V20CredProblemReport(
                description={
                    "en": "oh no",
                    "code": ProblemReportReason.ISSUANCE_ABANDONED.value,
                }
            )
            handler = test_module.CredProblemReportHandler()
            responder = MockResponder()
            await handler.handle(request_context, responder)

        mock_cred_mgr.assert_called_once_with(request_context.profile)
        mock_cred_mgr.return_value.receive_problem_report.assert_called_once_with(
            request_context.message, request_context.connection_record.connection_id
        )
        assert not responder.messages

    async def test_called_not_ready(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.connection_record = mock.MagicMock()
        request_context.connection_ready = False

        request_context.message = V20CredProblemReport(
            description={
                "en": "Change of plans",
                "code": ProblemReportReason.ISSUANCE_ABANDONED.value,
            }
        )
        handler = test_module.CredProblemReportHandler()
        responder = MockResponder()

        with self.assertRaises(test_module.HandlerException) as err:
            await handler.handle(request_context, responder)
        assert (
            err.exception.message
            == "Connection used for credential problem report not ready"
        )

    async def test_called_no_connection(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.connection_record = None

        request_context.message = V20CredProblemReport(
            description={
                "en": "Change of plans",
                "code": ProblemReportReason.ISSUANCE_ABANDONED.value,
            }
        )
        handler = test_module.CredProblemReportHandler()
        responder = MockResponder()

        with self.assertRaises(test_module.HandlerException) as err:
            await handler.handle(request_context, responder)
        assert (
            err.exception.message
            == "Connectionless not supported for credential problem report"
        )

        assert not responder.messages



"""Schemas for dif presentation exchange attachment."""

from typing import Mapping, Optional, Sequence

from marshmallow import (
    EXCLUDE,
    INCLUDE,
    ValidationError,
    fields,
    post_dump,
    pre_load,
    validate,
)

from ....messaging.models.base import BaseModel, BaseModelSchema
from ....messaging.valid import UUID4_EXAMPLE, UUID4_VALIDATE, StrOrNumberField
from ....vc.vc_ld.models.presentation import (
    VerifiablePresentation,
    VerifiablePresentationSchema,
)


class ClaimFormat(BaseModel):
    """Defines Claim field."""

    class Meta:
        """ClaimFormat metadata."""

        schema_class = "ClaimFormatSchema"

    def __init__(
        self,
        *,
        jwt: Optional[Mapping] = None,
        jwt_vc: Optional[Mapping] = None,
        jwt_vp: Optional[Mapping] = None,
        ldp: Optional[Mapping] = None,
        ldp_vc: Optional[Mapping] = None,
        ldp_vp: Optional[Mapping] = None,
        di_vc: Optional[Mapping] = None,
    ):
        """Initialize format."""
        self.jwt = jwt
        self.jwt_vc = jwt_vc
        self.jwt_vp = jwt_vp
        self.ldp = ldp
        self.ldp_vc = ldp_vc
        self.ldp_vp = ldp_vp
        self.di_vc = di_vc


class ClaimFormatSchema(BaseModelSchema):
    """Single ClaimFormat Schema."""

    class Meta:
        """ClaimFormatSchema metadata."""

        model_class = ClaimFormat
        unknown = EXCLUDE

    jwt = fields.Dict(required=False)
    jwt_vc = fields.Dict(required=False)
    jwt_vp = fields.Dict(required=False)
    ldp = fields.Dict(required=False)
    ldp_vc = fields.Dict(required=False)
    ldp_vp = fields.Dict(required=False)
    di_vc = fields.Dict(required=False)


class SubmissionRequirements(BaseModel):
    """describes input to be submitted via a presentation submission."""

    class Meta:
        """SubmissionRequirements metadata."""

        schema_class = "SubmissionRequirementsSchema"

    def __init__(
        self,
        *,
        _name: Optional[str] = None,
        purpose: Optional[str] = None,
        rule: Optional[str] = None,
        count: Optional[int] = None,
        minimum: Optional[int] = None,
        maximum: Optional[int] = None,
        _from: Optional[str] = None,
        # Self_reference
        from_nested: Optional[Sequence] = None,
    ):
        """Initialize SubmissionRequirement."""
        self._name = _name
        self.purpose = purpose
        self.rule = rule
        self.count = count
        self.minimum = minimum
        self.maximum = maximum
        self._from = _from
        self.from_nested = from_nested


class SubmissionRequirementsSchema(BaseModelSchema):
    """Single Presentation Definition Schema."""

    class Meta:
        """SubmissionRequirementsSchema metadata."""

        model_class = SubmissionRequirements
        unknown = EXCLUDE

    _name = fields.Str(required=False, data_key="name", metadata={"description": "Name"})
    purpose = fields.Str(required=False, metadata={"description": "Purpose"})
    rule = fields.Str(
        required=False,
        validate=validate.OneOf(["all", "pick"]),
        metadata={"description": "Selection"},
    )
    count = fields.Int(
        required=False,
        metadata={"description": "Count Value", "example": 1234, "strict": True},
    )
    minimum = fields.Int(
        required=False,
        data_key="min",
        metadata={"description": "Min Value", "example": 1234, "strict": True},
    )
    maximum = fields.Int(
        required=False,
        data_key="max",
        metadata={"description": "Max Value", "example": 1234, "strict": True},
    )
    _from = fields.Str(required=False, data_key="from", metadata={"description": "From"})
    # Self References
    from_nested = fields.List(
        fields.Nested(lambda: SubmissionRequirementsSchema()), required=False
    )

    @pre_load
    def validate_from(self, data, **kwargs):
        """Support validation of from and from_nested."""
        if "from" in data and "from_nested" in data:
            raise ValidationError(
                "Both from and from_nested cannot be "
                "specified in the submission requirement"
            )
        if "from" not in data and "from_nested" not in data:
            raise ValidationError(
                "Either from or from_nested needs to be "
                "specified in the submission requirement"
            )
        return data


class SchemaInputDescriptor(BaseModel):
    """SchemaInputDescriptor."""

    class Meta:
        """SchemaInputDescriptor metadata."""

        schema_class = "SchemaInputDescriptorSchema"

    def __init__(
        self,
        *,
        uri: Optional[str] = None,
        required: Optional[bool] = None,
    ):
        """Initialize SchemaInputDescriptor."""
        self.uri = uri
        self.required = required


class SchemaInputDescriptorSchema(BaseModelSchema):
    """Single SchemaField Schema."""

    class Meta:
        """SchemaInputDescriptorSchema metadata."""

        model_class = SchemaInputDescriptor
        unknown = EXCLUDE

    uri = fields.Str(required=False, metadata={"description": "URI"})
    required = fields.Bool(required=False, metadata={"description": "Required"})


class SchemasInputDescriptorFilter(BaseModel):
    """SchemasInputDescriptorFilter."""

    class Meta:
        """InputDescriptor Schemas filter metadata."""

        schema_class = "SchemasInputDescriptorFilterSchema"

    def __init__(
        self,
        *,
        oneof_filter: bool = False,
        uri_groups: Sequence[Sequence[SchemaInputDescriptor]] = None,
    ):
        """Initialize SchemasInputDescriptorFilter."""
        self.oneof_filter = oneof_filter
        self.uri_groups = uri_groups


class SchemasInputDescriptorFilterSchema(BaseModelSchema):
    """Single SchemasInputDescriptorFilterSchema Schema."""

    class Meta:
        """SchemasInputDescriptorFilterSchema metadata."""

        model_class = SchemasInputDescriptorFilter
        unknown = EXCLUDE

    uri_groups = fields.List(fields.List(fields.Nested(SchemaInputDescriptorSchema)))
    oneof_filter = fields.Bool(metadata={"description": "oneOf"})

    @pre_load
    def extract_info(self, data, **kwargs):
        """Deserialize."""
        new_data = {}
        if isinstance(data, dict):
            if "uri_groups" in data:
                return data
            elif "oneof_filter" in data and isinstance(data["oneof_filter"], list):
                new_data["oneof_filter"] = True
                uri_group_list_of_list = []
                uri_group_list = data.get("oneof_filter")
                for uri_group in uri_group_list:
                    if isinstance(uri_group, list):
                        uri_group_list_of_list.append(uri_group)
                    else:
                        uri_group_list_of_list.append([uri_group])
                new_data["uri_groups"] = uri_group_list_of_list
        elif isinstance(data, list):
            new_data["oneof_filter"] = False
            new_data["uri_groups"] = [data]
        data = new_data
        return data


class DIFHolder(BaseModel):
    """Single Holder object for Constraints."""

    class Meta:
        """Holder metadata."""

        schema_class = "DIFHolderSchema"

    def __init__(
        self,
        *,
        field_ids: Sequence[str] = None,
        directive: Optional[str] = None,
    ):
        """Initialize Holder."""
        self.field_ids = field_ids
        self.directive = directive


class DIFHolderSchema(BaseModelSchema):
    """Single Holder Schema."""

    class Meta:
        """DIFHolderSchema metadata."""

        model_class = DIFHolder
        unknown = EXCLUDE

    field_ids = fields.List(
        fields.Str(
            required=False,
            validate=UUID4_VALIDATE,
            metadata={"description": "FieldID", "example": UUID4_EXAMPLE},
        ),
        required=False,
        data_key="field_id",
    )
    directive = fields.Str(
        required=False,
        validate=validate.OneOf(["required", "preferred"]),
        metadata={"description": "Preference"},
    )


class Filter(BaseModel):
    """Single Filter for the Constraint object."""

    class Meta:
        """Filter metadata."""

        schema_class = "FilterSchema"

    def __init__(
        self,
        *,
        _not: bool = False,
        _type: Optional[str] = None,
        fmt: Optional[str] = None,
        pattern: Optional[str] = None,
        minimum: Optional[str] = None,
        maximum: Optional[str] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        exclusive_min: Optional[str] = None,
        exclusive_max: Optional[str] = None,
        const: Optional[str] = None,
        enums: Sequence[str] = None,
    ):
        """Initialize Filter."""
        self._type = _type
        self.fmt = fmt
        self.pattern = pattern
        self.minimum = minimum
        self.maximum = maximum
        self.min_length = min_length
        self.max_length = max_length
        self.exclusive_min = exclusive_min
        self.exclusive_max = exclusive_max
        self.const = const
        self.enums = enums
        self._not = _not


class FilterSchema(BaseModelSchema):
    """Single Filter Schema."""

    class Meta:
        """FilterSchema metadata."""

        model_class = Filter
        unknown = EXCLUDE

    _type = fields.Str(required=False, data_key="type", metadata={"description": "Type"})
    fmt = fields.Str(
        required=False, data_key="format", metadata={"description": "Format"}
    )
    pattern = fields.Str(required=False, metadata={"description": "Pattern"})
    minimum = StrOrNumberField(required=False, metadata={"description": "Minimum"})
    maximum = StrOrNumberField(required=False, metadata={"description": "Maximum"})
    min_length = fields.Int(
        required=False,
        data_key="minLength",
        metadata={"description": "Min Length", "example": 1234, "strict": True},
    )
    max_length = fields.Int(
        required=False,
        data_key="maxLength",
        metadata={"description": "Max Length", "example": 1234, "strict": True},
    )
    exclusive_min = StrOrNumberField(
        required=False,
        data_key="exclusiveMinimum",
        metadata={"description": "ExclusiveMinimum"},
    )
    exclusive_max = StrOrNumberField(
        required=False,
        data_key="exclusiveMaximum",
        metadata={"description": "ExclusiveMaximum"},
    )
    const = StrOrNumberField(required=False, metadata={"description": "Const"})
    enums = fields.List(
        StrOrNumberField(required=False, metadata={"description": "Enum"}),
        required=False,
        data_key="enum",
    )
    _not = fields.Boolean(
        required=False,
        data_key="not",
        metadata={"description": "Not", "example": False},
    )

    @pre_load
    def extract_info(self, data, **kwargs):
        """Enum validation and not filter logic."""
        if "not" in data:
            new_data = {"not": True}
            for key, value in data.get("not").items():
                new_data[key] = value
            data = new_data
        if "enum" in data:
            if not isinstance(data.get("enum"), list):
                raise ValidationError("enum is not specified as a list")
        return data

    @post_dump
    def serialize_reformat(self, data, **kwargs):
        """Support serialization of not filter according to DIF spec."""
        if data.pop("not", False):
            return {"not": data}

        return data


class DIFField(BaseModel):
    """Single Field object for the Constraint."""

    class Meta:
        """Field metadata."""

        schema_class = "DIFFieldSchema"

    def __init__(
        self,
        *,
        id: Optional[str] = None,
        paths: Sequence[str] = None,
        purpose: Optional[str] = None,
        predicate: Optional[str] = None,
        _filter: Optional[Filter] = None,
    ):
        """Initialize Field."""
        self.paths = paths
        self.purpose = purpose
        self.predicate = predicate
        self._filter = _filter
        self.id = id


class DIFFieldSchema(BaseModelSchema):
    """Single Field Schema."""

    class Meta:
        """DIFFieldSchema metadata."""

        model_class = DIFField
        unknown = EXCLUDE

    id = fields.Str(required=False, metadata={"description": "ID"})
    paths = fields.List(
        fields.Str(required=False, metadata={"description": "Path"}),
        required=False,
        data_key="path",
    )
    purpose = fields.Str(required=False, metadata={"description": "Purpose"})
    predicate = fields.Str(
        required=False,
        validate=validate.OneOf(["required", "preferred"]),
        metadata={"description": "Preference"},
    )
    _filter = fields.Nested(FilterSchema, data_key="filter")


class Constraints(BaseModel):
    """Single Constraints which describes InputDescriptor's Constraint field."""

    class Meta:
        """Constraints metadata."""

        schema_class = "ConstraintsSchema"

    def __init__(
        self,
        *,
        subject_issuer: Optional[str] = None,
        limit_disclosure: Optional[bool] = None,
        holders: Sequence[DIFHolder] = None,
        _fields: Sequence[DIFField] = None,
        status_active: Optional[str] = None,
        status_suspended: Optional[str] = None,
        status_revoked: Optional[str] = None,
    ):
        """Initialize Constraints for Input Descriptor."""
        self.subject_issuer = subject_issuer
        self.limit_disclosure = limit_disclosure
        self.holders = holders
        self._fields = _fields
        self.status_active = status_active
        self.status_suspended = status_suspended
        self.status_revoked = status_revoked


class ConstraintsSchema(BaseModelSchema):
    """Single Constraints Schema."""

    class Meta:
        """ConstraintsSchema metadata."""

        model_class = Constraints
        unknown = EXCLUDE

    subject_issuer = fields.Str(
        required=False,
        validate=validate.OneOf(["required", "preferred"]),
        data_key="subject_is_issuer",
        metadata={"description": "SubjectIsIssuer"},
    )
    limit_disclosure = fields.Str(
        required=False, metadata={"description": "LimitDisclosure"}
    )
    holders = fields.List(
        fields.Nested(DIFHolderSchema), required=False, data_key="is_holder"
    )
    _fields = fields.List(
        fields.Nested(DIFFieldSchema), required=False, data_key="fields"
    )
    status_active = fields.Str(
        required=False, validate=validate.OneOf(["required", "allowed", "disallowed"])
    )
    status_suspended = fields.Str(
        required=False, validate=validate.OneOf(["required", "allowed", "disallowed"])
    )
    status_revoked = fields.Str(
        required=False, validate=validate.OneOf(["required", "allowed", "disallowed"])
    )

    @pre_load
    def extract_info(self, data, **kwargs):
        """Support deserialization of statuses according to DIF spec."""
        if "statuses" in data:
            if "active" in data.get("statuses"):
                if "directive" in data.get("statuses").get("active"):
                    data["status_active"] = data["statuses"]["active"]["directive"]
            if "suspended" in data.get("statuses"):
                if "directive" in data.get("statuses").get("suspended"):
                    data["status_suspended"] = data["statuses"]["suspended"]["directive"]
            if "revoked" in data.get("statuses"):
                if "directive" in data.get("statuses").get("revoked"):
                    data["status_revoked"] = data["statuses"]["revoked"]["directive"]
        return data

    @post_dump
    def reformat_data(self, data, **kwargs):
        """Support serialization of statuses according to DIF spec."""
        if "status_active" in data:
            statuses = data.get("statuses", {})
            statuses["active"] = {"directive": data.get("status_active")}
            data["statuses"] = statuses
            del data["status_active"]
        if "status_suspended" in data:
            statuses = data.get("statuses", {})
            statuses["suspended"] = {"directive": data.get("status_suspended")}
            data["statuses"] = statuses
            del data["status_suspended"]
        if "status_revoked" in data:
            statuses = data.get("statuses", {})
            statuses["revoked"] = {"directive": data.get("status_revoked")}
            data["statuses"] = statuses
            del data["status_revoked"]
        return data


class InputDescriptors(BaseModel):
    """Input Descriptors."""

    class Meta:
        """InputDescriptors metadata."""

        schema_class = "InputDescriptorsSchema"

    def __init__(
        self,
        *,
        id: Optional[str] = None,
        groups: Sequence[str] = None,
        name: Optional[str] = None,
        purpose: Optional[str] = None,
        metadata: Optional[dict] = None,
        constraint: Optional[Constraints] = None,
        schemas: Optional[SchemasInputDescriptorFilter] = None,
    ):
        """Initialize InputDescriptors."""
        self.id = id
        self.groups = groups
        self.name = name
        self.purpose = purpose
        self.metadata = metadata
        self.constraint = constraint
        self.schemas = schemas


class InputDescriptorsSchema(BaseModelSchema):
    """Single InputDescriptors Schema."""

    class Meta:
        """InputDescriptorsSchema metadata."""

        model_class = InputDescriptors
        unknown = EXCLUDE

    id = fields.Str(required=False, metadata={"description": "ID"})
    groups = fields.List(
        fields.Str(required=False, metadata={"description": "Group"}),
        required=False,
        data_key="group",
    )
    name = fields.Str(required=False, metadata={"description": "Name"})
    purpose = fields.Str(required=False, metadata={"description": "Purpose"})
    metadata = fields.Dict(
        required=False, metadata={"description": "Metadata dictionary"}
    )
    constraint = fields.Nested(ConstraintsSchema, required=False, data_key="constraints")
    schemas = fields.Nested(
        SchemasInputDescriptorFilterSchema,
        required=False,
        data_key="schema",
        metadata={
            "description": (
                "Accepts a list of schema or a dict containing filters like oneof_filter."
            ),
            "example": {
                "oneof_filter": [
                    [
                        {"uri": "https://www.w3.org/Test1#Test1"},
                        {"uri": "https://www.w3.org/Test2#Test2"},
                    ],
                    {
                        "oneof_filter": [
                            [{"uri": "https://www.w3.org/Test1#Test1"}],
                            [{"uri": "https://www.w3.org/Test2#Test2"}],
                        ]
                    },
                ]
            },
        },
    )


class Requirement(BaseModel):
    """Single Requirement generated from toRequirement function."""

    class Meta:
        """Requirement metadata."""

        schema_class = "RequirementSchema"

    def __init__(
        self,
        *,
        count: Optional[int] = None,
        maximum: Optional[int] = None,
        minimum: Optional[int] = None,
        input_descriptors: Sequence[InputDescriptors] = None,
        nested_req: Optional[Sequence] = None,
    ):
        """Initialize Requirement."""
        self.count = count
        self.maximum = maximum
        self.minimum = minimum
        self.input_descriptors = input_descriptors
        self.nested_req = nested_req


class RequirementSchema(BaseModelSchema):
    """Single Requirement Schema."""

    class Meta:
        """RequirementSchema metadata."""

        model_class = Requirement
        unknown = EXCLUDE

    count = fields.Int(
        required=False,
        metadata={"description": "Count Value", "example": 1234, "strict": True},
    )
    maximum = fields.Int(
        required=False,
        metadata={"description": "Max Value", "example": 1234, "strict": True},
    )
    minimum = fields.Int(
        required=False,
        metadata={"description": "Min Value", "example": 1234, "strict": True},
    )
    input_descriptors = fields.List(fields.Nested(InputDescriptorsSchema), required=False)
    # Self References
    nested_req = fields.List(
        fields.Nested(lambda: RequirementSchema(exclude=("_nested_req",))),
        required=False,
    )


class PresentationDefinition(BaseModel):
    """https://identity.foundation/presentation-exchange/."""

    class Meta:
        """PresentationDefinition metadata."""

        schema_class = "PresentationDefinitionSchema"

    def __init__(
        self,
        *,
        id: Optional[str] = None,
        name: Optional[str] = None,
        purpose: Optional[str] = None,
        fmt: Optional[ClaimFormat] = None,
        submission_requirements: Sequence[SubmissionRequirements] = None,
        input_descriptors: Sequence[InputDescriptors] = None,
        **kwargs,
    ):
        """Initialize flattened single-JWS to include in attach decorator data."""
        super().__init__(**kwargs)
        self.id = id
        self.name = name
        self.purpose = purpose
        self.fmt = fmt
        self.submission_requirements = submission_requirements
        self.input_descriptors = input_descriptors


class PresentationDefinitionSchema(BaseModelSchema):
    """Single Presentation Definition Schema."""

    class Meta:
        """PresentationDefinitionSchema metadata."""

        model_class = PresentationDefinition
        unknown = EXCLUDE

    id = fields.Str(
        required=False,
        validate=UUID4_VALIDATE,
        metadata={
            "description": "Unique Resource Identifier",
            "example": UUID4_EXAMPLE,
        },
    )
    name = fields.Str(
        required=False,
        metadata={
            "description": (
                "Human-friendly name that describes what the presentation definition"
                " pertains to"
            )
        },
    )
    purpose = fields.Str(
        required=False,
        metadata={
            "description": (
                "Describes the purpose for which the Presentation Definition's inputs"
                " are being requested"
            )
        },
    )
    fmt = fields.Nested(ClaimFormatSchema, required=False, data_key="format")
    submission_requirements = fields.List(
        fields.Nested(SubmissionRequirementsSchema), required=False
    )
    input_descriptors = fields.List(fields.Nested(InputDescriptorsSchema), required=False)


class InputDescriptorMapping(BaseModel):
    """Single InputDescriptorMapping object."""

    class Meta:
        """InputDescriptorMapping metadata."""

        schema_class = "InputDescriptorMappingSchema"

    def __init__(
        self,
        *,
        id: Optional[str] = None,
        fmt: Optional[str] = None,
        path: Optional[str] = None,
    ):
        """Initialize InputDescriptorMapping."""
        self.id = id
        self.fmt = fmt
        self.path = path


class InputDescriptorMappingSchema(BaseModelSchema):
    """Single InputDescriptorMapping Schema."""

    class Meta:
        """InputDescriptorMappingSchema metadata."""

        model_class = InputDescriptorMapping
        unknown = EXCLUDE

    id = fields.Str(required=False, metadata={"description": "ID"})
    fmt = fields.Str(
        required=False,
        dump_default="ldp_vc",
        data_key="format",
        metadata={"description": "Format"},
    )
    path = fields.Str(required=False, metadata={"description": "Path"})


class PresentationSubmission(BaseModel):
    """Single PresentationSubmission object."""

    class Meta:
        """PresentationSubmission metadata."""

        schema_class = "PresentationSubmissionSchema"

    def __init__(
        self,
        *,
        id: Optional[str] = None,
        definition_id: Optional[str] = None,
        descriptor_maps: Sequence[InputDescriptorMapping] = None,
    ):
        """Initialize InputDescriptorMapping."""
        self.id = id
        self.definition_id = definition_id
        self.descriptor_maps = descriptor_maps


class PresentationSubmissionSchema(BaseModelSchema):
    """Single PresentationSubmission Schema."""

    class Meta:
        """PresentationSubmissionSchema metadata."""

        model_class = PresentationSubmission
        unknown = EXCLUDE

    id = fields.Str(
        required=False,
        metadata={"description": "ID", "example": UUID4_EXAMPLE},
    )
    definition_id = fields.Str(
        required=False,
        metadata={"description": "DefinitionID", "example": UUID4_EXAMPLE},
    )
    descriptor_maps = fields.List(
        fields.Nested(InputDescriptorMappingSchema),
        required=False,
        data_key="descriptor_map",
    )


class VPWithSubmission(VerifiablePresentation):
    """Single VerifiablePresentation object."""

    class Meta:
        """VerifiablePresentation metadata."""

        schema_class = "VPWithSubmissionSchema"

    def __init__(
        self,
        *,
        presentation_submission: Optional[PresentationSubmission] = None,
        **kwargs,
    ):
        """Initialize VerifiablePresentation."""
        super().__init__(**kwargs)
        self.presentation_submission = presentation_submission


class VPWithSubmissionSchema(VerifiablePresentationSchema):
    """Single Verifiable Presentation Schema."""

    class Meta:
        """VerifiablePresentationSchema metadata."""

        model_class = VPWithSubmission
        unknown = INCLUDE

    presentation_submission = fields.Nested(PresentationSubmissionSchema)


class DIFOptions(BaseModel):
    """Single DIFOptions object."""

    class Meta:
        """DIFOptions metadata."""

        schema_class = "DIFOptionsSchema"

    def __init__(
        self,
        *,
        challenge: Optional[str] = None,
        domain: Optional[str] = None,
    ):
        """Initialize DIFOptions."""
        self.challenge = challenge
        self.domain = domain


class DIFOptionsSchema(BaseModelSchema):
    """Schema for options required for the Prover to fulfil the Verifier's request."""

    class Meta:
        """DIFOptionsSchema metadata."""

        model_class = DIFOptions
        unknown = EXCLUDE

    challenge = fields.String(
        required=False,
        validate=UUID4_VALIDATE,
        metadata={
            "description": "Challenge protect against replay attack",
            "example": UUID4_EXAMPLE,
        },
    )
    domain = fields.String(
        required=False,
        metadata={
            "description": "Domain protect against replay attack",
            "example": "4jt78h47fh47",
        },
    )
